import numpy as np  # librería para manejos de arreglos
import pandas as pd  # librería para el tratamiento de datos
from sklearn.cluster import KMeans

from sklearn.preprocessing import MinMaxScaler
# from sklearn.metrics import silhouette_samples
from tslearn.barycenters import euclidean_barycenter, dtw_barycenter_averaging, softdtw_barycenter
from scipy import cluster, spatial


class Conglomerados:
    """
    Clase que contiene métodos para realizar los cluster permite metodos de DTW y K-means estan definidos en esta clase
    """

    List_clusterTree = []  # Lista vacía para almacenar dataframe obtenidos de los cluster
    Curva_mediaTree = []  # Lista vacía para almacenar
    List_Electric_Feature = []  # lista vacia para almacenar dataframe de datos electricos
    Cluster_NumberTree = int  # entero para almacenar la cantidad máxima de clustlista_num = []
    lista_Nu_cluster = []  # lista para almacenar número de cluster, la posicion coicide con el número de cluster
    indexprimercluster = []
    list_residuos = []  # lista para alamacenar los residuos entre la curva DBA y los perfiles de carga
    list_residuos_media = []  # Lista para alamcenar valores medios de los residuos de cada cluster
    corr_cluster = []  # Lista para almacenar valores de la correlación de cada cluster
    corr_cluster_media = []  # Lista para almacenar la medai de los valores de correlación total de cada cluster


    #print("Análisis de Curvas de Carga")

    def __init__(self, escenario, stdist=None, season=None, typeDay=None):
        self.escenario = escenario
        self.stdist = stdist
        self.season = season
        self.typeDay = typeDay

    def normalice_axis(self, df):
        """"Método que normaliza datasets frame por axisas."""
        # df = df.dropna(how ='any')
        scaler = MinMaxScaler(feature_range=(0, 1))
        scaled_df = scaler.fit_transform(df.transpose())
        scaled_df = pd.DataFrame(scaled_df.T)
        scaled_df = scaled_df.set_index(df.index)
        return scaled_df

    def normalice(self, df):
        """Método para normalizar"""
        scaler = MinMaxScaler(feature_range=(0, 1))
        scaled_df = scaler.fit_transform(df)
        scaled_df = pd.DataFrame(scaled_df)
        return scaled_df

    def DTWDistance(self, s1, s2):
        """Método para calcular distancía dinámica temporal"""

        DTW = {}

        for i in range(len(s1)):
            DTW[(i, -1)] = float('inf')
        for i in range(len(s2)):
            DTW[(-1, i)] = float('inf')
        DTW[(-1, -1)] = 0

        for i in range(len(s1)):
            for j in range(len(s2)):
                dist = (s1[i] - s2[j]) ** 2
                DTW[(i, j)] = dist + min(DTW[(i - 1, j)], DTW[(i, j - 1)], DTW[(i - 1, j - 1)])

        return np.sqrt(DTW[len(s1) - 1, len(s2) - 1])

    def ClusterIndicesNumpy(self, clustNum, labels_array):  # función para agrupar las observaciones por cluster
        """Métodos para agrupara observaciones"""
        return np.where(labels_array == clustNum)[0]

    def ClusterIndicesComp(self, clustNum, labels_array):  # list comprehension
        return np.array([i for i, x in enumerate(labels_array) if x == clustNum])

    def clusterizar(self):
        """#  Métodos para clusterizar"""
        escenario_ = self.escenario.dropna(how='any')
        E_H = np.array(escenario_)
        E_W = np.array(self.normalice_axis(escenario_))  # array normalizado algoritmo Normalizado dataframe de energías para DTW
        df_List_W = self.normalice_axis(escenario_)  # datafreme normalizado para poder normalizar DTW y graficar.

        f = lambda u, v: np.sqrt(((u - v) ** 2).sum())  # distancia euclidiana.
        k = lambda S1, S2: self.DTWDistance(S1, S2)  # distancia DTW.

        Array_DTW = cluster.hierarchy.fclusterdata(E_W, t=1.7, criterion='distance', metric=k, depth=2, method='ward', R=None)  # lista que muestra en cada posición el numero de cluster al que pertenece
        # arrayTree_ = self.ClusterIndicesNumpy(1,Array_DTW)    # devuelve un arreglo con las posiciones de las observaciones pertenecientes al cluster 3
        self.Cluster_NumberTree = Array_DTW.max()  # contiene el número máximo de cluster obtenidos
        self.lista_Nu_cluster = self.listaCluster(self.Cluster_NumberTree)

        for i in range(1, self.Cluster_NumberTree + 1):
            arrayTree = self.ClusterIndicesNumpy(i, Array_DTW)  # devuelve un arreglo con las posiciones de las observaciones pertenecientes al cluster i

            # arrayTree = ClusterIndicesComp(Cluster_NumberTree, Array_DTW)
            # Y[ClusterIndicesNumpy(2,kmeans.labels_)]                          #devuelve el vector de las dimensiones de las observaciones que hay en cada cluster
            # print(arrayTree)
            # df_Temp = df_List.ix[arrayTree]                                   # filtra el datasets frame segun la lista array
            df_Temp = df_List_W.iloc[arrayTree]  # filtra el datasets frame segun la lista array nrmalizada se obteiene un datasets frame con los perfiles de un cluster
            df_Temp_E = self.stdist.iloc[arrayTree]  # dataframe que contiene valores estadisticos de cada perfil individual
            # C_Media = df_Temp.mean()                                          # se calcula el promedio de las curvas dentro de un mismo cluster
            Z = np.array(df_Temp)  # se crea un NumpyArray del dataframe que contiene el clsuter de la iteración i.
            dba_Media = dtw_barycenter_averaging(Z, max_iter=100, verbose=False)  # se realiza la media de DTW.
            # dba_Media = softdtw_barycenter(Z, gamma=1., max_iter=100)
            dba_Media = pd.DataFrame(dba_Media)
            dba_Media_scalaed = self.normalice(dba_Media)
            self.List_clusterTree.append(df_Temp)
            self.List_Electric_Feature.append(df_Temp_E)
            self.Curva_mediaTree.append(dba_Media_scalaed)  # contiene un arreglo de vectores que representan la curva de

        for Q in range(0, len(self.List_clusterTree)):
            media_ = self.Curva_mediaTree[Q].T
            residuos = pd.DataFrame(
                self.List_clusterTree[Q].values - media_.values,
                columns=self.List_clusterTree[Q].columns
            )  # calculos de los residuos
            ECM = sum(residuos.pow(2).mean()) / len(self.List_clusterTree[Q])  # calcula los errores cuadraticos mnedios respecta a la media DBA
            ECM = round(ECM, 3)
            self.list_residuos.append(ECM)

            list_corr_value = []
            temp_matrix = self.List_clusterTree[Q]
            temp_matrix = temp_matrix[~np.all(temp_matrix == 0, axis=1)]  # elimina filas con todos los elemetos cero
            for b in range(0, len(temp_matrix)):
                # curva_carga = self.List_clusterTree[Q].iloc[b].values
                curva_carga = temp_matrix.iloc[b].values
                curva_tipica = self.Curva_mediaTree[Q].T.values
                correlacion = self.norm_cross_corr(curva_carga, curva_tipica)
                correlacion = round(correlacion, 4)
            list_corr_value.append(correlacion)

            self.corr_cluster.append(list_corr_value)
            self.corr_cluster_media.append(sum(list_corr_value) / len(list_corr_value))

    def norm_cross_corr(self, serie1, serie2):
        """ La función para encontrar la correlación cruzada normalizada en python es: statsmodels.tsa.stattools.ccf
            :param: serie1: serie de tiempo\n
            :param: serie2: serie de tiempo promedio\n
            :return: valor de correlacion entre las series"""
        return np.sum(serie1 * serie2) / (np.linalg.norm(serie1) * np.linalg.norm(serie2))

    """
    def listResiduos(self,Curva_mediaTree,List_clusterTree):
        Metodo que permite calcular el error cuadratico medio a entre entre el vector de curva promedios y las restantes curvas dentro de grupo.\n\n
        :param : Curva_mediaTree: lista que continen en cada posicion la curva promedio. \n
        :param : List_clusterTree: lista que contiene en cada posición los gruppos de curva 
                resultado de clustering 

        for Q in range(0, len(List_clusterTree)):
            media_ = Curva_mediaTree[Q].T
            residuos = pd.DataFrame(List_clusterTree[Q].values - media_.values, columns=List_clusterTree[Q].columns)
            # EC = residuos.pow(2)
            ECM = sum(residuos.pow(2).mean()) / len(List_clusterTree[Q])
            ECM = round(ECM, 3)
            self.list_residuos.append(ECM)
    """

    def Kmeans_Cluster(self, df, Cluster_Number, Normalice=0):
        """Parametros datafreme a clsuterizar, nuemro de lcuster, 1 si normaliza por filas y cero si normaliza por columnas.\n
        parameter:Cluster_Number \n
        parameter:Normalice"""

        List_cluster = []  # Lista vacía para almacenar
        # Curva_media = []
        if Normalice == 1:
            df_nor = self.normalice_axis(df)
        else:
            df_nor = self.normalice(df)
        # df_nor = normalice_axis(df)# normmaliza datasets frame de enrgias
        array_df = np.array(df_nor)
        kmeans = KMeans(n_clusters=Cluster_Number, init='k-means++', max_iter=300).fit(
            array_df)  # pasamos como parámetros la lsita de puntos X.

        # centroids = kmeans.cluster_centers_  # nos devuelve los centroirde de los clúster
        # Predicting the clusters
        # labels = kmeans.predict(array_df)  # devuelve una lista alineada que nos dice a que clúster pertenece cada punto
        # num_cluster_points = kmeans.labels_.tolist()  # muestra el numero de cluster separados por ,

        for i in range(0, Cluster_Number):
            array = self.ClusterIndicesNumpy(i, kmeans.labels_)  # devuelve un arreglo con las posiciones de las observaciones pertenecientes al cluster 3
            df_Temp = df_nor.iloc[array]  # fitra el datasets frame segun la lista array
            List_cluster.append(df_Temp)  # almacena en un arreglo los dataframe dque representa cada cluster.
        return List_cluster

    def listaCluster(self, num):
        """
        Metodo para almacenar en una lista los numero de cluster\n
        :argument : num\n
        :return: lista de los cluster
        """
        lista_num = list(range(num))
        for i in lista_num:
            lista_num[i] = str(lista_num[i] + 1)

        return lista_num


# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm
from openpy_fxts.baseline_mdls import base_class


class RNN_Dense_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='RNN_Dense',
            type_mdl='RNN',
            app='mdits'
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl, app)
        return



# -*- coding: utf-8 -*-
# @Time    : 29/05/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm


import tensorflow as tf

tkc = tf.keras.callbacks
tku = tf.keras.utils


def learning_curve(history):
    import matplotlib.pyplot as plt

    plt.style.use('seaborn-dark-palette')
    plt.figure(figsize=(17, 7))
    plt.plot(history.history["loss"], label="training")
    plt.plot(history.history["val_loss"], label="validation")
    plt.xlabel("Epoch")
    plt.ylabel("Loss -Mean Squared Error")
    plt.legend()
    plt.show()


def callbacks(source, folder, name_model):
    import datetime

    callbacks_list = []

    log_dir = "logs\\" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    tensorboard_callback = tkc.TensorBoard(log_dir=log_dir, histogram_freq=1)
    callbacks_list.append(tensorboard_callback)
    # w_fn = '.\\architectures\\electricity_prediction-1-{}.h5'.format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))

    early_stopping = tkc.EarlyStopping(
        monitor='val_loss',
        patience=5,
        verbose=1,
        mode='auto'
    )
    callbacks_list.append(early_stopping)
    # Change the cooldown to 1, if behances unexpectedly
    learning_rate_reduction = tkc.ReduceLROnPlateau(
        monitor='val_loss',
        patience=5,
        verbose=1,
        factor=0.2,
        min_lr=2.5e-5,
        cooldown=0
    )
    callbacks_list.append(learning_rate_reduction)
    file_path = source + '/' + folder + '/' + f"{name_model}" + "/weights-{epoch:02d}.hdf5"
    checkpoint_cb = tkc.ModelCheckpoint(
        filepath=file_path,
        monitor='val_loss',
        verbose=1,
        save_best_only=True,
        save_weights_only=True,
        mode='auto'
    )
    callbacks_list.append(checkpoint_cb)
    return callbacks_list



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, value_miss


class gru2_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(gru2_dense, self).__init__()
        if config is None:
            config = {
                'gru1': {
                    'units': 256,
                    'activation': 'relu'
                },
                'gru2': {
                    'units': 256,
                    'activation': 'relu'
                },
                'dropout': 0.3,
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.gru1 = tkl.GRU(
            units=config['gru1']['units'],
            return_sequences=True
        )
        self.drop1 = tkl.Dropout(config['dropout'])
        self.gru2 = tkl.GRU(
            units=config['gru2']['units'],
            activation=config['gru2']['activation']
        )
        self.drop2 = tkl.Dropout(config['dropout'])
        self.dense = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.reshape = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.gru1(x)
        x = self.drop1(x)
        x = self.gru2(x)
        x = self.drop2(x)
        x = self.dense(x)
        x = self.reshape(x)
        return x



# -*- coding: utf-8 -*-
# @Time    : 25/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm


import torch
import numpy as np
from scipy import optimize
import pandas as pd


class generate_missing_data:

    def __init__(
            self,
            seed: int = 123
    ):
        self.seed = seed

    # Function produce_NA for generating missing values ------------------------------------------------------
    def produce_NaN(
            self,
            df_true: pd.DataFrame = None,
            p_miss: float = 0.40,
            mecha: str = "MCAR",
            opt="logistic",
            p_obs=0.05,
            q=0.3
    ):

        """
        Generate missing values for specifics missing-datasets mechanism and proportion of missing values.

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data for which missing values will be simulated.
            If a numpy array is provided, it will be converted to a pytorch tensor.
        p_miss : float
            Proportion of missing values to generate for variables which will have missing values.
        mecha : str,
                Indicates the missing-datasets mechanism to be used. "MCAR" by default, "MAR", "MNAR" or "Mask"
        opt: str,
                For mecha = "MNAR", it indicates how the missing-datasets mechanism is generated: using a logistic regression ("logistic"), quantile censorship ("quantile") or logistic regression for generating a self-masked MNAR mechanism ("selfmasked").
        p_obs : float
                If mecha = "MAR", or mecha = "MNAR" with opt = "logistic" or "quanti", proportion of variables with *no* missing values that will be used for the logistic masking model.
        q : float
            If mecha = "MNAR" and opt = "quanti", quantile level at which the cuts should occur.

        Returns
        ----------
        A dictionnary containing:
        'X_init': the initial datasets matrix.
        'X_incomp': the datasets with the generated missing values.
        'mask': a matrix indexing the generated missing values.s
        :param df_true:
        """

        X = df_true.copy()
        X = X.to_numpy()
        to_torch = torch.is_tensor(X)  ## output a pytorch tensor, or a numpy array
        if not to_torch:
            X = X.astype(np.float32)
            X = torch.from_numpy(X)

        if mecha == "MAR":
            mask = generate_missing_data(self.seed)._MAR_mask(X, p_miss, p_obs).double()
        elif mecha == "MNAR" and opt == "logistic":
            mask = generate_missing_data(self.seed)._MNAR_mask_logistic(X, p_miss, p_obs).double()
        elif mecha == "MNAR" and opt == "quantile":
            mask = generate_missing_data(self.seed)._MNAR_mask_quantiles(X, p_miss, q, 1 - p_obs).double()
        elif mecha == "MNAR" and opt == "selfmasked":
            mask = generate_missing_data(self.seed)._MNAR_self_mask_logistic(X, p_miss).double()
        else:
            mask = (torch.rand(X.shape) < p_miss).double()

        X_nas = X.clone()
        X_nas[mask.bool()] = np.nan
        X_nas.double()


        df_data_miss = pd.DataFrame(X_nas, columns=df_true.columns)
        df_data_miss['datetime'] = df_true.index
        df_data_miss.set_index('datetime', inplace=True)

        return df_data_miss

    def produce_NaN_pattern(self, df_data: pd.DataFrame = None, columns: list = None, prob_missing: float = 0.3, Random: bool = False):

        if columns is None:
            columns = list(df_data.columns)
        df_data = df_data.copy()
        np.random.seed(self.seed)
        W_train = np.random.choice(2, df_data[columns].shape, p=[prob_missing, 1 - prob_missing])
        W_train = pd.DataFrame(W_train, columns=columns).replace(0, np.nan)
        W_train['datetime'] = df_data.index
        W_train.set_index('datetime', inplace=True)
        if not Random:
            aux = list(W_train.columns)
            aux2 = aux[1:]
            for i in aux2:
                W_train[i] = W_train[aux[0]]
        X_train = (df_data[columns] * W_train)
        df_data[columns] = X_train
        return df_data

    def _nanmean(self, v, *args, **kwargs):
        """
        A Pytorch version on Numpy's nanmean
        """
        v = v.clone()
        is_nan = torch.isnan(v)
        v[is_nan] = 0
        return v.sum(*args, **kwargs) / (~is_nan).float().sum(*args, **kwargs)

    #### Quantile ######
    def _quantile(self, X, q, dim=None):
        """
        Returns the q-th quantile.

        Parameters
        ----------
        X : torch.DoubleTensor or torch.cuda.DoubleTensor, shape (n, d)
            Input datasets.

        q : float
            Quantile level (starting from lower values).

        dim : int or None, default = None
            Dimension allong which to compute quantiles. If None, the tensor is flattened and one value is returned.


        Returns
        -------
            quantiles : torch.DoubleTensor

        """
        return X.kthvalue(int(q * len(X)), dim=dim)[0]

    #### Automatic selection of the regularization parameter ####
    def _pick_epsilon(self, X, quant=0.5, mult=0.05, max_points=2000):
        """
            Returns a quantile (times a multiplier) of the halved pairwise squared distances in X.
            Used to select a regularization parameter for Sinkhorn distances.

        Parameters
        ----------
        X : torch.DoubleTensor or torch.cuda.DoubleTensor, shape (n, d)
            Input datasets on which distances will be computed.

        quant : float, default = 0.5
            Quantile to return (default is median).

        mult : float, default = 0.05
            Mutiplier to apply to the quantiles.

        max_points : int, default = 2000
            If the length of X is larger than max_points, estimate the quantile on a random subset of size max_points to
            avoid memory overloads.

        Returns
        -------
            epsilon: float

        """
        means = generate_missing_data(self.seed)._nanmean(X, 0)
        X_ = X.clone()
        mask = torch.isnan(X_)
        X_[mask] = (mask * means)[mask]

        np.random.seed(self.seed)
        idx = np.random.choice(len(X_), min(max_points, len(X_)), replace=False)
        X = X_[idx]
        dists = ((X[:, None] - X) ** 2).sum(2).flatten() / 2.
        dists = dists[dists > 0]

        return generate_missing_data(self.seed)._quantile(dists, quant, 0).item() * mult

    #### Accuracy Metrics ####
    def _MAE(self, X, X_true, mask):
        """
        Mean Absolute Error (MAE) between imputed variables and ground truth. Pytorch/Numpy agnostic

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data with imputed variables.

        X_true : torch.DoubleTensor or np.ndarray, shape (n, d)
            Ground truth.

        mask : torch.BoolTensor or np.ndarray of booleans, shape (n, d)
            Missing value mask (missing if True)

        Returns
        -------
            MAE : float

        """
        if torch.is_tensor(mask):
            mask_ = mask.bool()
            return torch.abs(X[mask_] - X_true[mask_]).sum() / mask_.sum()
        else:  # should be an ndarray
            mask_ = mask.astype(bool)
            return np.absolute(X[mask_] - X_true[mask_]).sum() / mask_.sum()

    def _RMSE(self, X, X_true, mask):
        """
        Root Mean Squared Error (MAE) between imputed variables and ground truth. Pytorch/Numpy agnostic

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data with imputed variables.

        X_true : torch.DoubleTensor or np.ndarray, shape (n, d)
            Ground truth.

        mask : torch.BoolTensor or np.ndarray of booleans, shape (n, d)
            Missing value mask (missing if True)

        Returns
        -------
            RMSE : float

        """
        if torch.is_tensor(mask):
            mask_ = mask.bool()
            return (((X[mask_] - X_true[mask_]) ** 2).sum() / mask_.sum()).sqrt()
        else:  # should be an ndarray
            mask_ = mask.astype(bool)
            return np.sqrt(((X[mask_] - X_true[mask_]) ** 2).sum() / mask_.sum())

    ##################### MISSING DATA MECHANISMS #############################

    ##### Missing At Random ######

    def _MAR_mask(self, X, p, p_obs):
        """
        Missing at random mechanism with a logistic masking model. First, a subset of variables with *no* missing values is
        randomly selected. The remaining variables have missing values according to a logistic model with random weights,
        re-scaled so as to attain the desired proportion of missing values on those variables.

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data for which missing values will be simulated. If a numpy array is provided,
            it will be converted to a pytorch tensor.

        p : float
            Proportion of missing values to generate for variables which will have missing values.

        p_obs : float
            Proportion of variables with *no* missing values that will be used for the logistic masking model.

        Returns
        -------
        mask : torch.BoolTensor or np.ndarray (depending on type of X)
            Mask of generated missing values (True if the value is missing).

        """

        n, d = X.shape

        to_torch = torch.is_tensor(X)  ## output a pytorch tensor, or a numpy array
        if not to_torch:
            X = torch.from_numpy(X)

        mask = torch.zeros(n, d).bool() if to_torch else np.zeros((n, d)).astype(bool)

        d_obs = max(int(p_obs * d), 1)  ## number of variables that will have no missing values (at least one variable)
        d_na = d - d_obs  ## number of variables that will have missing values

        ### Sample variables that will all be observed, and those with missing values:
        np.random.seed(self.seed)
        idxs_obs = np.random.choice(d, d_obs, replace=False)
        idxs_nas = np.array([i for i in range(d) if i not in idxs_obs])

        ### Other variables will have NA proportions that depend on those observed variables, through a logistic model
        ### The parameters of this logistic model are random.

        ### Pick coefficients so that W^Tx has unit variance (avoids shrinking)
        coeffs = generate_missing_data(self.seed)._pick_coeffs(X, idxs_obs, idxs_nas)
        ### Pick the intercepts to have a desired amount of missing values
        intercepts = generate_missing_data(self.seed)._fit_intercepts(X[:, idxs_obs], coeffs, p)

        ps = torch.sigmoid(X[:, idxs_obs].mm(coeffs) + intercepts)

        ber = torch.rand(n, d_na)
        mask[:, idxs_nas] = ber < ps

        return mask

    ##### Missing not at random ######

    def _MNAR_mask_logistic(self, X, p, p_params=.3, exclude_inputs=True):
        """
        Missing not at random mechanism with a logistic masking model. It implements two mechanisms:
        (i) Missing probabilities are selected with a logistic model, taking all variables as inputs. Hence, values that are
        inputs can also be missing.
        (ii) Variables are split into a set of intputs for a logistic model, and a set whose missing probabilities are
        determined by the logistic model. Then inputs are then masked MCAR (hence, missing values from the second set will
        depend on masked values.
        In either case, weights are random and the intercept is selected to attain the desired proportion of missing values.

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data for which missing values will be simulated.
            If a numpy array is provided, it will be converted to a pytorch tensor.

        p : float
            Proportion of missing values to generate for variables which will have missing values.

        p_params : float
            Proportion of variables that will be used for the logistic masking model (only if exclude_inputs).

        exclude_inputs : boolean, default=True
            True: mechanism (ii) is used, False: (i)

        Returns
        -------
        mask : torch.BoolTensor or np.ndarray (depending on type of X)
            Mask of generated missing values (True if the value is missing).

        """

        n, d = X.shape

        to_torch = torch.is_tensor(X)  ## output a pytorch tensor, or a numpy array
        if not to_torch:
            X = torch.from_numpy(X)
        mask = torch.zeros(n, d).bool() if to_torch else np.zeros((n, d)).astype(bool)
        d_params = max(int(p_params * d), 1) if exclude_inputs else d  ## number of variables used as inputs (at least 1)
        d_na = d - d_params if exclude_inputs else d  ## number of variables masked with the logistic model

        ### Sample variables that will be parameters for the logistic regression:
        np.random.seed(self.seed)
        idxs_params = np.random.choice(d, d_params, replace=False) if exclude_inputs else np.arange(d)
        idxs_nas = np.array([i for i in range(d) if i not in idxs_params]) if exclude_inputs else np.arange(d)

        ### Other variables will have NA proportions selected by a logistic model
        ### The parameters of this logistic model are random.

        ### Pick coefficients so that W^Tx has unit variance (avoids shrinking)
        coeffs = generate_missing_data(self.seed)._pick_coeffs(X, idxs_params, idxs_nas)
        ### Pick the intercepts to have a desired amount of missing values
        intercepts = generate_missing_data(self.seed)._fit_intercepts(X[:, idxs_params], coeffs, p)

        ps = torch.sigmoid(X[:, idxs_params].mm(coeffs) + intercepts)

        ber = torch.rand(n, d_na)
        mask[:, idxs_nas] = ber < ps

        ## If the inputs of the logistic model are excluded from MNAR missingness,
        ## mask some values used in the logistic model at random.
        ## This makes the missingness of other variables potentially dependent on masked values

        if exclude_inputs:
            mask[:, idxs_params] = torch.rand(n, d_params) < p

        return mask

    def _MNAR_self_mask_logistic(self, X, p):
        """
        Missing not at random mechanism with a logistic self-masking model. Variables have missing values probabilities
        given by a logistic model, taking the same variable as input (hence, missingness is independent from one variable
        to another). The intercepts are selected to attain the desired missing rate.

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data for which missing values will be simulated.
            If a numpy array is provided, it will be converted to a pytorch tensor.

        p : float
            Proportion of missing values to generate for variables which will have missing values.

        Returns
        -------
        mask : torch.BoolTensor or np.ndarray (depending on type of X)
            Mask of generated missing values (True if the value is missing).

        """

        n, d = X.shape

        to_torch = torch.is_tensor(X)  ## output a pytorch tensor, or a numpy array
        if not to_torch:
            X = torch.from_numpy(X)

        ### Variables will have NA proportions that depend on those observed variables, through a logistic model
        ### The parameters of this logistic model are random.

        ### Pick coefficients so that W^Tx has unit variance (avoids shrinking)
        coeffs = generate_missing_data(self.seed)._pick_coeffs(X, self_mask=True)
        ### Pick the intercepts to have a desired amount of missing values
        intercepts = generate_missing_data(self.seed)._fit_intercepts(X, coeffs, p, self_mask=True)

        ps = torch.sigmoid(X * coeffs + intercepts)
        np.random.seed(self.seed)
        ber = torch.rand(n, d) if to_torch else np.random.rand(n, d)
        mask = ber < ps if to_torch else ber < ps.numpy()

        return mask

    def _MNAR_mask_quantiles(self, X, p, q, p_params, cut='both', MCAR=False):
        """
        Missing not at random mechanism with quantile censorship. First, a subset of variables which will have missing
        variables is randomly selected. Then, missing values are generated on the q-quantiles at random. Since
        missingness depends on quantile information, it depends on masked values, hence this is a MNAR mechanism.

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data for which missing values will be simulated.
            If a numpy array is provided, it will be converted to a pytorch tensor.

        p : float
            Proportion of missing values to generate for variables which will have missing values.

        q : float
            Quantile level at which the cuts should occur

        p_params : float
            Proportion of variables that will have missing values

        cut : 'both', 'upper' or 'lower', default = 'both'
            Where the cut should be applied. For instance, if q=0.25 and cut='upper', then missing values will be generated
            in the upper quartiles of selected variables.

        MCAR : bool, default = True
            If true, masks variables that were not selected for quantile censorship with a MCAR mechanism.

        Returns
        -------
        mask : torch.BoolTensor or np.ndarray (depending on type of X)
            Mask of generated missing values (True if the value is missing).

        """
        n, d = X.shape

        to_torch = torch.is_tensor(X)  ## output a pytorch tensor, or a numpy array
        if not to_torch:
            X = torch.from_numpy(X)

        mask = torch.zeros(n, d).bool() if to_torch else np.zeros((n, d)).astype(bool)

        d_na = max(int(p_params * d), 1)  ## number of variables that will have NMAR values

        ### Sample variables that will have imps at the extremes
        np.random.seed(self.seed)
        idxs_na = np.random.choice(d, d_na, replace=False)  ### select at least one variable with missing values

        ### check if values are greater/smaller that corresponding quantiles
        if cut == 'upper':
            quants = generate_missing_data(self.seed)._quantile(X[:, idxs_na], 1 - q, dim=0)
            m = X[:, idxs_na] >= quants
        elif cut == 'lower':
            quants = generate_missing_data(self.seed)._quantile(X[:, idxs_na], q, dim=0)
            m = X[:, idxs_na] <= quants
        elif cut == 'both':
            u_quants = generate_missing_data(self.seed)._quantile(X[:, idxs_na], 1 - q, dim=0)
            l_quants = generate_missing_data(self.seed)._quantile(X[:, idxs_na], q, dim=0)
            m = (X[:, idxs_na] <= l_quants) | (X[:, idxs_na] >= u_quants)

        ### Hide some values exceeding quantiles
        ber = torch.rand(n, d_na)
        mask[:, idxs_na] = (ber < p) & m

        if MCAR:
            ## Add a mcar mecanism on top
            mask = mask | (torch.rand(n, d) < p)

        return mask

    def _pick_coeffs(self, X, idxs_obs=None, idxs_nas=None, self_mask=False):
        n, d = X.shape
        if self_mask:
            coeffs = torch.randn(d)
            Wx = X * coeffs
            coeffs /= torch.std(Wx, 0)
        else:
            d_obs = len(idxs_obs)
            d_na = len(idxs_nas)
            coeffs = torch.randn(d_obs, d_na)
            Wx = X[:, idxs_obs].mm(coeffs)
            coeffs /= torch.std(Wx, 0, keepdim=True)
        return coeffs

    def _fit_intercepts(self, X, coeffs, p, self_mask=False):
        if self_mask:
            d = len(coeffs)
            intercepts = torch.zeros(d)
            for j in range(d):
                def f(x):
                    return torch.sigmoid(X * coeffs[j] + x).mean().item() - p

                intercepts[j] = optimize.bisect(f, -50, 50)
        else:
            d_obs, d_na = coeffs.shape
            intercepts = torch.zeros(d_na)
            for j in range(d_na):
                def f(x):
                    return torch.sigmoid(X.mv(coeffs[:, j]) + x).mean().item() - p

                intercepts[j] = optimize.bisect(f, -50, 50)
        return intercepts






import numpy as np
from numba import jit

from scipy.stats import mode
from scipy.spatial.distance import squareform
from sklearn.metrics import accuracy_score

#from utils.progbar import progress_bar

try:
    from IPython.core.display import HTML, Javascript, display
except ImportError:
    pass


__all__ = ['dtw_distance', 'KnnDTW']


def slow_dtw_distance(series1, series2):
    """
        Returns the DTW similarity distance between two 1-D
        timeseries numpy arrays.
        Arguments:
            series1, series2 : array of shape [n_timepoints]
                Two arrays containing n_samples of timeseries datasets
                whose DTW distance between each sample of A and B
                will be compared
        Returns:
            DTW distance between sequence 1 and 2
        """
    l1 = series1.shape[0]
    l2 = series2.shape[0]
    E = np.empty((l1, l2))

    # Fill First Cell
    E[0][0] = np.square(series1[0] - series2[0])

    # Fill First Column
    for i in range(1, l1):
        E[i][0] = E[i - 1][0] + np.square(series1[i] - series2[0])

    # Fill First Row
    for i in range(1, l2):
        E[0][i] = E[0][i - 1] + np.square(series1[0] - series2[i])

    for i in range(1, l1):
        for j in range(1, l2):
            v = np.square(series1[i] - series2[j])

            v1 = E[i - 1][j]
            v2 = E[i - 1][j - 1]
            v3 = E[i][j - 1]

            if v1 <= v2 and v1 <= v3:
                E[i][j] = v1 + v
            elif v2 <= v1 and v2 <= v3:
                E[i][j] = v2 + v
            else:
                E[i][j] = v3 + v

    return np.sqrt(E[-1][-1])


@jit(nopython=True)
def dtw_distance(series1, series2):
    """
    Returns the DTW similarity distance between two 2-D
    timeseries numpy arrays.
    Arguments:
        series1, series2 : array of shape [n_timepoints]
            Two arrays containing n_samples of timeseries datasets
            whose DTW distance between each sample of A and B
            will be compared
    Returns:
        DTW distance between sequence 1 and 2
    """
    l1 = series1.shape[0]
    l2 = series2.shape[0]
    E = np.empty((l1, l2))

    # Fill First Cell
    E[0][0] = np.square(series1[0] - series2[0])

    # Fill First Column
    for i in range(1, l1):
        E[i][0] = E[i - 1][0] + np.square(series1[i] - series2[0])

    # Fill First Row
    for i in range(1, l2):
        E[0][i] = E[0][i - 1] + np.square(series1[0] - series2[i])

    for i in range(1, l1):
        for j in range(1, l2):
            v = np.square(series1[i] - series2[j])

            v1 = E[i - 1][j]
            v2 = E[i - 1][j - 1]
            v3 = E[i][j - 1]

            if v1 <= v2 and v1 <= v3:
                E[i][j] = v1 + v
            elif v2 <= v1 and v2 <= v3:
                E[i][j] = v2 + v
            else:
                E[i][j] = v3 + v

    return np.sqrt(E[-1][-1])


# Modified from https://github.com/markdregan/K-Nearest-Neighbors-with-Dynamic-Time-Warping
class KnnDTW(object):
    """K-nearest neighbor classifier using dynamic time warping
    as the distance measure between pairs of time series arrays
    Arguments
    ---------
    n_neighbors : int, optional (default = 1)
        Number of neighbors to use by default for KNN
    """

    def __init__(self, n_neighbors=1):
        self.n_neighbors = n_neighbors

    def fit(self, x, y):
        """Fit the model using x as training datasets and y as class labels
        Arguments
        ---------
        x : array of shape [n_samples, n_timepoints]
            Training datasets set for input into KNN classifer
        y : array of shape [n_samples]
            Training labels for input into KNN classifier
        """

        self.x = np.copy(x)
        self.y = np.copy(y)

    def _dist_matrix(self, x, y):
        """Computes the M x N distance matrix between the training
        dataset and testing dataset (y) using the DTW distance measure
        Arguments
        ---------
        x : array of shape [n_samples, n_timepoints]
        y : array of shape [n_samples, n_timepoints]
        Returns
        -------
        Distance matrix between each item of x and y with
            shape [training_n_samples, testing_n_samples]
        """

        # Compute the distance matrix
        dm_count = 0

        # Compute condensed distance matrix (upper triangle) of pairwise dtw distances
        # when x and y are the same array
        if (np.array_equal(x, y)):
            x_s = np.shape(x)
            dm = np.zeros((x_s[0] * (x_s[0] - 1)) // 2, dtype=np.double)

            #p = progress_bar(np.shape(dm)[0])

            for i in range(0, x_s[0] - 1):
                for j in range(i + 1, x_s[0]):
                    dm[dm_count] = dtw_distance(x[i], y[j])

                    dm_count += 1
                    #p.update(dm_count)

            # Convert to squareform
            dm = squareform(dm)
            return dm

        # Compute full distance matrix of dtw distnces between x and y
        else:
            x_s = np.shape(x)
            y_s = np.shape(y)
            dm = np.zeros((x_s[0], y_s[0]))
            dm_size = x_s[0] * y_s[0]

            #p = progress_bar(dm_size)

            for i in range(0, x_s[0]):
                for j in range(0, y_s[0]):
                    dm[i, j] = dtw_distance(x[i], y[j])
                    # Update progress bar
                    dm_count += 1
                    #p.update(dm_count)

            return dm

    def predict(self, x):
        """Predict the class labels or probability estimates for
        the provided datasets
        Arguments
        ---------
          x : array of shape [n_samples, n_timepoints]
              Array containing the testing datasets set to be classified
        Returns
        -------
          2 arrays representing:
              (1) the predicted class labels
              (2) the knn label count probability
        """
        np.random.seed(0)
        dm = self._dist_matrix(x, self.x)

        # Identify the k nearest neighbors
        knn_idx = dm.argsort()[:, :self.n_neighbors]

        # Identify k nearest labels
        knn_labels = self.y[knn_idx]

        # Model Label
        mode_data = mode(knn_labels, axis=1)
        mode_label = mode_data[0]
        mode_proba = mode_data[1] / self.n_neighbors

        return mode_label.ravel(), mode_proba.ravel()

    def evaluate(self, x, y):
        """
        Predict the class labels or probability estimates for
        the provided datasets and then evaluates the accuracy score.
        Arguments
        ---------
          x : array of shape [n_samples, n_timepoints]
              Array containing the testing datasets set to be classified
          y : array of shape [n_samples]
              Array containing the labels of the testing dataset to be classified
        Returns
        -------
          1 floating point value representing the accuracy of the classifier
        """
        # Predict the labels and the probabilities
        pred_labels, pred_probas = self.predict(x)

        # Ensure labels are integers
        y = y.astype('int32')
        pred_labels = pred_labels.astype('int32')

        # Compute accuracy measure
        accuracy = accuracy_score(y, pred_labels)
        return accuracy

    def predict_proba(self, x):
        """Predict the class labels probability estimates for
        the provided datasets
        Arguments
        ---------
            x : array of shape [n_samples, n_timepoints]
                Array containing the testing datasets set to be classified
        Returns
        -------
            2 arrays representing:
                (1) the predicted class probabilities
                (2) the knn labels
        """
        np.random.seed(0)
        dm = self._dist_matrix(x, self.x)

        # Invert the distance matrix
        dm = -dm

        # Compute softmax probabilities
        dm_exp = np.exp(dm - dm.max())
        dm = dm_exp / np.sum(dm_exp, axis=-1, keepdims=True)

        classes = np.unique(self.y)
        class_dm = []

        # Partition distance matrix by class
        for i, cls in enumerate(classes):
            idx = np.argwhere(self.y == cls)[:, 0]
            cls_dm = dm[:, idx]  # [N_test, N_train_c]

            # Take maximum distance vector due to softmax probabilities
            cls_dm = np.max(cls_dm, axis=-1)  # [N_test,]

            class_dm.append([cls_dm])

        # Concatenate the classwise distance matrices and transpose
        class_dm = np.concatenate(class_dm, axis=0)  # [C, N_test]
        class_dm = class_dm.transpose()  # [N_test, C]

        # Compute softmax probabilities
        class_dm_exp = np.exp(class_dm - class_dm.max())
        class_dm = class_dm_exp / np.sum(class_dm_exp, axis=-1, keepdims=True)

        probabilities = class_dm
        knn_labels = np.argmax(probabilities, axis=-1)

        return probabilities, knn_labels


# -*- coding: utf-8 -*-
# @Time    : 25/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
from sklearn.impute import SimpleImputer
from openpy_fxts.models.imputation.utilities.Utils import Metrics, Utils
from openpy_fxts.models.imputation.base_lib import init_models
from openpy_fxts.utils import get_numeric_categorical
import impyute as impy
import numpy as np



mt = Metrics()


class imp_basic(init_models):

    def __init__(
            self,
            df_miss: pd.DataFrame = None,
            df_true: pd.DataFrame = None,
            metrics: dict = None
    ):
        super().__init__(df_miss, df_true)
        self.df_miss = df_miss
        self.df_true = df_true
        self.metrics = metrics
        self.list_aux = Utils(self.df_miss).list_column_missing()

    def constant(
            self,
            method: list
    ):
        """
        
        :param method: 
        :return: 
        """
        # imputing with a constan
        df_imp = self.df_miss.copy(deep=True)
        df_imp.iloc[:, :] = SimpleImputer(strategy=method).fit_transform(df_imp)
        if self.df_true is None:
            return df_imp
        else:
            if self.metrics is None or len(self.metrics) == 0:
                self.metrics = dict()
            self.metrics = mt.add_dict_metrics(
                self.df_true[self.list_aux],
                df_imp[self.list_aux],
                self.metrics,
                f'const_{method}'
            )
            return df_imp, self.metrics

    def fillna(
            self,
            method: list = None,
            aux: str = 'mean'
    ):
        """
        :param list_fillna:
        :return:
        """
        df_imp = self.df_miss.copy(deep=True)
        df_imp.fillna(method=method, inplace=True)
        if Utils(df_imp).check_missing():
            df_imp.iloc[:, :] = SimpleImputer(strategy=aux).fit_transform(df_imp)
        if self.df_true is None:
            return df_imp
        else:
            if self.metrics is None or len(self.metrics) == 0:
                self.metrics = dict()
            self.metrics = mt.add_dict_metrics(
                self.df_true[self.list_aux],
                df_imp[self.list_aux],
                self.metrics,
                f'const_{method}'
            )
            return df_imp, self.metrics

    def moving_window_replace(
            self,
            res_min: int = 15,
            n_days: int = None,
            previous: bool = None,
            back: bool = None
    ):
        df_imp = self.df_miss.copy(deep=True)
        data = df_imp.values
        if previous is None and back is None:
            previous = True
        window = (24 * 60) / res_min
        if n_days is not None:
            n_days = 1
        window = window * n_days
        n = 0
        window = int(window)
        while len(data[np.isnan(data)]) > 0:
            print(f'iter:{n}, n_NaN:{len(data[np.isnan(data)])}')
            for row in range(data.shape[0]):
                for col in range(data.shape[1]):
                    if np.isnan(data[row, col]):
                        if previous:
                            try:
                                data[row, col] = data[row - window, col]
                            except IndexError:
                                try:
                                    data[row, col] = data[row + window, col]
                                except IndexError:
                                    pass
                        if back:
                            try:
                                data[row, col] = data[row + window, col]
                            except IndexError:
                                try:
                                    data[row, col] = data[row - window, col]
                                except IndexError:
                                    pass
            n += 1
        df_imp.iloc[:, :] = data

        if self.df_true is None:
            return df_imp
        else:
            if self.metrics is None or len(self.metrics) == 0:
                self.metrics = dict()
            self.metrics = mt.add_dict_metrics(
                self.df_true[self.list_aux],
                df_imp[self.list_aux],
                self.metrics,
                'window_replace'
            )
            return df_imp, self.metrics

    def moving_window_mean(
            self,
            n_hours: int = None,
            res_min: int = None,
            n_past: bool = None,
            n_future: bool = None,
    ):
        df_imp = self.df_miss.copy(deep=True)
        data = df_imp.values
        aux = 0
        if n_hours is None:
            n = 1
        n_hours = n * 60
        while len(data[np.isnan(data)]) > 0:
            print(f'iter:{aux}, n_NaN:{len(data[np.isnan(data)])}')
            if n_past is None and n_future is None and res_min is not None:
                n_past, n_future = n_hours / res_min, n_hours / res_min
            elif n_past is None and n_future is None and res_min is None:
                n_past, n_future = 4, 4
            for row in range(data.shape[0]):
                for col in range(data.shape[1]):
                    if np.isnan(data[row, col]):
                        data[row, col] = np.concatenate(
                            (
                                data[row - n_past: row, col],
                                data[row: row + n_future, col]
                            )
                        ).mean()
                        if np.isnan(data[row, col]):
                            data[row, col] = data[row - n_past: row, col].mean()
                            if np.isnan(data[row, col]):
                                data[row, col] = data[row: row + n_future, col].mean()
            aux += 1
        df_imp.iloc[:, :] = data
        if self.df_true is None:
            return df_imp
        else:
            if self.metrics is None or len(self.metrics) == 0:
                self.metrics = dict()
            self.metrics = mt.add_dict_metrics(
                self.df_true[self.list_aux],
                df_imp[self.list_aux],
                self.metrics,
                'window_mean'
            )
            return df_imp, self.metrics

    def interpolate(
            self,
            method: str,
            axis=0,
            limit=None,
            inplace=True,
            limit_direction=None,
            limit_area=None,
            downcast=None,
            order=2,
            aux: str = 'mean'
    ):
        """
        :param method:
        :param axis:
        :param limit:
        :param inplace:
        :param limit_direction:
        :param limit_area:
        :param downcast:
        :param order:
        :param aux:
        :return:
        """
        df_imp = self.df_miss.copy(deep=True)
        df_imp.interpolate(
            method=method,
            axis=axis,
            limit=limit,
            inplace=inplace,
            limit_direction=limit_direction,
            limit_area=limit_area,
            downcast=downcast,
            order=order
        )
        if Utils(df_imp).check_missing():
            model = SimpleImputer(strategy=aux)
            df_imp.iloc[:, :] = model.fit_transform(df_imp)
        if self.df_true is None:
            return df_imp
        else:
            if self.metrics is None or len(self.metrics) == 0:
                self.metrics = dict()
            self.metrics = mt.add_dict_metrics(
                self.df_true[self.list_aux],
                df_imp[self.list_aux],
                self.metrics,
                f'const_{method}'
            )
            return df_imp, self.metrics



from .cluster_stage import Conglomerados
from .load_files import MultiplesArchivos
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
import numpy as np  # libreria para arreglos
from matplotlib import cm
import pandas as pd  # Librería para para tratamiento de datos, (dataframe)
import glob
import os


class Mult_escenarios:

    dfList = []  # Declara una lista vacía
    dfList_User = []
    Ener = pd.DataFrame()
    df_Season = pd.DataFrame()
    df_Season_Stadistic = pd.DataFrame()
    noFile = int
    progress = int

    def loadFiles(self, path, season, typeDay):

        self.path = path
        self.season = season
        self.typeDay = typeDay

        self.Estudio = MultiplesArchivos(
            self.path,
            self.season,
            self.typeDay
        )  # Se crea el objeto

        os.chdir(self.path)  # cambia el directorio de trabajo a la direcion Indir.
        filelist = glob.glob("*.xlsx")  # Muestra una lista de todos los archivos excel dentro del directorio
        i = 0
        for filename in filelist:
            print(filename)
            i = i + 1
            data_user = pd.read_excel(filename)
            id_users = self.Estudio.user(data_user)  # obtener el terminal de usuario(el ID)
            data_user = self.Estudio.cleanExcel(data_user)  # Limpiar el excel
            data_user = self.Estudio.proces_DF(data_user)  # Desagregar las fechas en columnas por separado.
            data_user = self.Estudio.fil_type_day(data_user)  # Filtramos dataframe por tipo de día de las semanas
            Ener_mean = self.Estudio.maxDeman(data_user, id_users)  # Máxima demanda consumida en el escenario
            self.Ener = pd.concat([self.Estudio.Ener, Ener_mean], axis=1)  # se añade la fila de energías obtenidas
            df_season = self.Estudio.fil_season(data_user)  # Se obtiene el datasets frame filtrado por los meses de estación
            df_Ener_season = self.Estudio.ener_hora_acum(df_season)  # datasets frame de energía por horas
            df_Ener_vector = self.Estudio.prom_clean(df_Ener_season, id_users)  # se obtiene promedio indexado por ID de usuario
            self.Estudio.df_Season = pd.concat([self.Estudio.df_Season, df_Ener_vector], axis=0)  # se concatenan los datos ciclicamente.
            df_Estadistic_Season = self.Estudio.stadistic(df_Ener_vector, df_season)
            self.Estudio.df_Season_Stadistic = pd.concat([self.Estudio.df_Season_Stadistic, df_Estadistic_Season], axis=0)
            print('here')

    def madeCluster(self):
        self.Escenario = Conglomerados(
            self.Estudio.df_Season,
            self.Estudio.df_Season_Stadistic,
            self.season,
            self.typeDay
        )  # Se crea el objeto
        self.Escenario.clusterizar()





from openpy_fxts.preprocessing_data.forecast.prepare_data import pre_processing_data
from openpy_fxts.architectures.utils import callbacks, learning_curve
from openpy_fxts.architectures.utils import _values_preliminary
from openpy_fxts.architectures.layers.layers_class import StackedSciNet
from openpy_fxts.base_tf import tkm, tkl, tku


class RNN_Stacked:

    def __init__(self, config_data=None, config_mdl=None):
        self.config_data = config_data
        self.n_past = config_data['n_past']
        self.n_future = config_data['n_future']
        self.n_inp_ft = config_data['n_inp_ft']
        self.n_out_ft = config_data['n_out_ft']
        # Parameters for model
        self.optimizer = config_mdl['optimizer']
        self.loss = config_mdl['loss']
        self.metrics = config_mdl['metrics']
        self.batch_size = config_mdl['batch_size']  # Batch size for training.
        self.epochs = config_mdl['epochs']  # Number of epochs to train for.
        self.units = config_mdl['units']  # no of lstm units
        self.dropout = config_mdl['dropout']
        self.name_model = 'LSTM_Stacked'

    def build_model(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft), name='Input')
        lstm1 = tkl.LSTM(
            64,
            activation='relu',
            return_sequences=True,
            input_shape=(self.n_past, self.n_inp_ft)
        )(input_layer)
        lstm2 = tkl.LSTM(64, activation='relu')(lstm1)
        dense1 = tkl.Dense(128)(lstm2)
        dropout = tkl.Dropout(0.3)(dense1)
        dense2 = tkl.Dense(self.n_future * self.n_out_ft, activation='relu')(dropout)
        output_layer = tkl.Reshape((self.n_future, self.n_out_ft))(dense2)
        output_layer = tkl.Conv1D(self.n_out_ft, 1, padding='same')(output_layer)
        # Connect input and output through the Model class
        model = tkm.Model(inputs=input_layer, outputs=output_layer, name='model')
        # Return the model
        return model

    def train_model(
            self,
            filepath: str = None,
            preliminary: bool = True
    ):
        data = pre_processing_data(self.config_data, train=True, valid=True)
        pre_processed = data.transformer_data()
        model = RNN_Stacked(self.config_data, self.config_mdl).build_model()
        model.compile(
            optimizer=self.optimizer,
            loss=self.loss,
            metrics=self.metrics
        )
        if self.config_data['view_summary']:
            model.summary()
        if self.config_data['plt_model']:
            if filepath is None:
                tku.plot_model(model, show_shapes=True)
            else:
                tku.plot_model(model, to_file=filepath, show_shapes=True, show_layer_names=True)
        # Training
        history = model.fit(
            pre_processed['train']['X'],
            pre_processed['train']['y'],
            epochs=self.epochs,
            validation_data=(
                pre_processed['valid']['X'],
                pre_processed['valid']['y']
            ),
            batch_size=self.batch_size,
            verbose=1,
            callbacks=callbacks(filepath, weights=True)
        )
        if self.config_data['plt_history']:
            learning_curve(history, self.name_model, self.config_data['time_init'], filepath=filepath)
        if preliminary:
            data = pre_processing_data(self.config_data, test=True)
            dict_test = data.transformer_data()
            _values_preliminary(model, dict_test, self.config_data)
        return model

    def prediction(
            self,
            model
    ):
        data = pre_processing_data(self.config_data, test=True)
        dict_test = data.transformer_data()
        yhat = _values_preliminary(model, dict_test, self.config_data)
        return yhat


class Stacked_SciNet:

    def __init__(self, config_data=None, config_mdl=None):
        self.config_data = config_data
        self.n_past = config_data['n_past']
        self.n_future = config_data['n_future']
        self.n_inp_ft = config_data['n_inp_ft']
        self.n_out_ft = config_data['n_out_ft']
        # Parameters for model
        self.config_mdl = config_mdl
        self.optimizer = config_mdl['optimizer']
        self.loss = config_mdl['loss']
        self.metrics = config_mdl['metrics']
        self.batch_size = config_mdl['batch_size']  # Batch size for training.
        self.epochs = config_mdl['epochs']  # Number of epochs to train for.
        self.units = config_mdl['units']  # no of lstm units
        self.dropout = config_mdl['dropout']
        self.name_model = 'Stacked_SciNet'

    def build_model(self):
        input_train = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        output_train = tkl.Input(shape=(self.n_future, self.n_out_ft))

        predictions = StackedSciNet(
            horizon=self.n_future,
            features=self.n_out_ft,
            stacks=2,
            levels=1,
            h=4,
            kernel_size=5
        )(input_train)
        sciNet_model = tkm.Model(inputs=input_train, outputs=predictions)

        # Return the model
        return sciNet_model

    def train_model(
            self,
            filepath: str = None,
            preliminary: bool = True
    ):
        data = pre_processing_data(self.config_data, train=True, valid=True)
        pre_processed = data.transformer_data()
        model = Stacked_SciNet(self.config_data, self.config_mdl).build_model()
        model.compile(
            optimizer=self.optimizer,
            loss=self.loss,
            metrics=self.metrics
        )
        if self.config_data['view_summary']:
            model.summary()
        if self.config_data['plt_model']:
            if filepath is None:
                tku.plot_model(model, show_shapes=True)
            else:
                tku.plot_model(model, to_file=filepath, show_shapes=True, show_layer_names=True)
        # Training
        history = model.fit(
            pre_processed['train']['X'],
            pre_processed['train']['y'],
            epochs=self.epochs,
            validation_data=(
                pre_processed['valid']['X'],
                pre_processed['valid']['y']
            ),
            batch_size=self.batch_size,
            verbose=1,
            callbacks=callbacks(filepath, weights=True)
        )
        if self.config_data['plt_history']:
            learning_curve(history, self.name_model, self.config_data['time_init'], filepath=filepath)
        if preliminary:
            data = pre_processing_data(self.config_data, test=True)
            dict_test = data.transformer_data()
            _values_preliminary(model, dict_test, self.config_data)
        return model

    def prediction(
            self,
            model
    ):
        data = pre_processing_data(self.config_data, test=True)
        dict_test = data.transformer_data()
        yhat = _values_preliminary(model, dict_test, self.config_data)
        return yhat


# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt
import datetime
from openpy_fxts.mdits.mdls_dl.base_tf import tko, tkloss, tku, tkc, tkmtc, tf
from time import time
from openpy_fxts.base_tf import value_miss


def tk_optimizer_loss(optimizer: str = 'Adam'):
    if optimizer == 'Adam':
        opt = tko.Adam()
        return opt


def tk_loss(loss: str = 'Huber'):
    if loss == 'Huber':
        metric = tkloss.Huber()
    return metric


def _date_init_final(config: dict = None):
    init = config['dataset'].index[0]
    final = config['dataset'].index[-1]
    date1 = f"{init.year}{init.month}{init.day}"
    date2 = f"{final.year}{final.month}{final.day}"
    n_in_out = f"int{config['n_past']}_out{config['n_future']}"
    fts = ""
    for i in config['y_colname']:
        fts += f"_{i}"
    name_folder = date1 + '_' + date2 + fts + '_' + n_in_out
    mttr_upd = 'mttr_' + config['MTTR'] + '_upd_' + config['Act']

    return name_folder, mttr_upd


def _learning_curve(
        history,
        name_mode: str = 'model',
        filepath: str = None,
        time_start=None,
        monitor_label: str = 'val_loss'
):
    hist = pd.DataFrame(history.history)
    hist['epoch'] = history.epoch
    min_vec = []
    for i in range(0, len(hist)):
        if i == 0:
            min_vec.append(hist[[monitor_label]].iloc[i, :].values[0])
        else:
            if hist[[monitor_label]].iloc[i, :].values[0] < np.min(min_vec):
                min_vec.append(hist[[monitor_label]].iloc[i, :].values[0])
            else:
                min_vec.append(np.min(min_vec))

    hist['best_from_start'] = min_vec
    hist['improved'] = np.where(hist['best_from_start'] == hist[monitor_label], 1, 0)

    best_epoch = hist[hist[monitor_label] == np.min(hist[monitor_label])]['epoch'].values[0]

    plt.figure(figsize=(18, 9))
    plt.xticks(np.arange(len(hist['epoch'])), np.arange(1, len(hist['epoch']) + 1))
    plt.plot(hist[monitor_label], label=monitor_label)
    try:
        plt.plot(hist['loss'], label='loss')
    except:
        pass
    # plt.plot(hist[monitor_label], label = 'val_loss')

    plt.scatter(
        hist[hist['improved'] == 1]['epoch'],
        hist[hist['improved'] == 1][monitor_label],
        color='green',
        label='Improved'
    )
    plt.legend()
    plt.title(
        'Learning curve'
        '\n' + 'best epoch:' + str(best_epoch + 1) + '(' + str(round(np.min(hist[monitor_label]), 3)) + ')' +
        '\n' + 'time_train: ' + str(datetime.timedelta(seconds=time() - time_start))
    )
    if filepath is not None:
        plt.savefig(f'{filepath}/{name_mode}_{monitor_label}.png')


def _mdl_characteristics(model, config_sim: dict = None, filepath: str = None):
    if config_sim['view_summary']:
        model.summary(expand_nested=True)
    if config_sim['plt_model']:
        if filepath is None:
            tku.plot_model(model, show_shapes=True)
        else:
            tku.plot_model(model, to_file=filepath, show_shapes=True, show_layer_names=True)


def _callbacks(
        filepath,
        model: bool = None,
        weights: bool = None,
        verbose: int = None,
        patience: int = None,
        model_checkpoint: bool = True,
        reduce_LR: bool = True,
        early_stopping: bool = True,
        tensor_board: bool = False
):
    if weights:
        save_best_only = True
        save_weights_only = True
        label = 'weights'
    if model:
        save_best_only = True
        save_weights_only = False
        label = 'model'

    callbacks_list = []
    if model_checkpoint:
        filepath = str(filepath) + f'/{label}' + '-{epoch:02d}.hdf5'
        checkpoint_cb = tkc.ModelCheckpoint(
            filepath=filepath,
            monitor='val_loss',
            verbose=verbose,
            save_best_only=save_best_only,
            save_weights_only=save_weights_only,
            mode='auto'
        )
        callbacks_list.append(checkpoint_cb)
    if reduce_LR:
        lr_cb = tkc.ReduceLROnPlateau(
            monitor='loss',
            factor=0.2,
            patience=patience,
            min_lr=2.5e-5,
            verbose=verbose,
            cooldown=0
        )
        callbacks_list.append(lr_cb)
    if early_stopping:
        es_cp = tkc.EarlyStopping(
            monitor='val_loss',
            patience=patience,
            verbose=verbose
        )
        callbacks_list.append(es_cp)
    if tensor_board:
        log_dir = "logs\\" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = tkc.TensorBoard(log_dir=log_dir, histogram_freq=1)
        callbacks_list.append(tensorboard_callback)
    return callbacks_list


def _scaler_data(train_dict, val_dict):
    scaler = StandardScaler()
    encoder_input_train = scaler.fit_transform(
        train_dict['encoder_input_train'].reshape(-1, train_dict['encoder_input_train'].shape[-1])).reshape(
        train_dict['encoder_input_train'].shape)
    decoder_input_train = scaler.transform(
        train_dict['decoder_input_train'].reshape(-1, train_dict['decoder_input_train'].shape[-1])).reshape(
        train_dict['decoder_input_train'].shape)
    final_output_train = scaler.transform(
        train_dict['final_output_train'].reshape(-1, train_dict['final_output_train'].shape[-1])).reshape(
        train_dict['final_output_train'].shape)
    encoder_input_val = scaler.transform(
        val_dict['encoder_input_val'].reshape(-1, val_dict['encoder_input_val'].shape[-1])).reshape(
        val_dict['encoder_input_val'].shape)
    decoder_input_val = scaler.transform(
        val_dict['decoder_input_val'].reshape(-1, val_dict['decoder_input_val'].shape[-1])).reshape(
        val_dict['decoder_input_val'].shape)
    final_output_val = scaler.transform(
        val_dict['final_output_val'].reshape(-1, val_dict['final_output_val'].shape[-1])).reshape(
        val_dict['final_output_val'].shape)

    train_dict_scaler, val_dict_scaler = {}, {}
    # Train
    train_dict_scaler['encoder_input_train'] = encoder_input_train
    train_dict_scaler['decoder_input_train'] = decoder_input_train
    train_dict_scaler['final_output_train'] = final_output_train
    # Validation
    val_dict_scaler['encoder_input_val'] = encoder_input_val
    val_dict_scaler['decoder_input_val'] = decoder_input_val
    val_dict_scaler['final_output_val'] = final_output_val

    return train_dict_scaler, val_dict_scaler, scaler


def _inverse_transform(yhat, dict_test, columns):
    for index, i in enumerate(columns):
        scaler = dict_test['scaler']['scaler_' + i]
        yhat[:, :, index] = scaler.inverse_transform(yhat[:, :, index])
        dict_test['y'][:, :, index] = scaler.inverse_transform(dict_test['y'][:, :, index])
    return yhat, dict_test['y']


def _historical_max_min(encoder_input_train):
    historical_max = np.expand_dims(
        np.max(encoder_input_train.reshape(-1, encoder_input_train.shape[-1]), axis=0, keepdims=True), 0)
    historical_min = np.expand_dims(
        np.min(encoder_input_train.reshape(-1, encoder_input_train.shape[-1]), axis=0, keepdims=True), 0)
    return historical_max, historical_min


def _example_plot(ax, final_output_test, pred, i, fontsize=10, hide_labels=False):
    ax.plot(final_output_test.reshape(-1, final_output_test.shape[-1])[:, i], c='blue', label="measured")
    ax.plot(pred.reshape(-1, pred.shape[-1])[:, i], c='red', label="prediction")
    ax.legend()
    ax.locator_params(nbins=3)
    if hide_labels:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
    else:
        ax.set_xlabel(f'Feature {i + 1} across all timesteps', fontsize=fontsize)
        ax.set_ylabel(f'F_{i + 1}', fontsize=fontsize)
        # ax.set_title('Title', fontsize=fontsize)
    mse = mean_squared_error(
        final_output_test.reshape(-1, final_output_test.shape[-1])[:, i],
        pred.reshape(-1, pred.shape[-1])[:, i],
        squared=False
    )
    print(f'Feature {i + 1} MSE: {mse}')


def _examples_plots_metrics_ft(
        ax,
        n_future,
        mae_list,
        type,
):
    ax.plot(range(0, n_future), mae_list, marker='o')
    ax.xticks((range(0, n_future)))
    ax.xlabel('Forecast Range')
    ax.ylabel(type)


def _examples_plots(
        ax,
        final_output_test,
        pred,
        i,
        fontsize,
        hide_labels=False
):
    ax.plot(final_output_test.reshape(-1, final_output_test.shape[-1])[:, i], c='blue', label="measured")
    ax.plot(pred.reshape(-1, pred.shape[-1])[:, i], c='red', label="prediction")
    ax.legend()
    ax.locator_params(nbins=3)
    if hide_labels:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
    else:
        ax.set_xlabel(f'Feature {i + 1} across all timesteps', fontsize=fontsize)
        ax.set_ylabel(f'F_{i + 1}', fontsize=fontsize)
        # ax.set_title('Title', fontsize=fontsize)
    mse = mean_squared_error(
        final_output_test.reshape(-1, final_output_test.shape[-1])[:, i],
        pred.reshape(-1, pred.shape[-1])[:, i],
        squared=False
    )
    print(f'Feature {i + 1} MSE: {mse}')


def _plot_results(
        final_output_test,
        pred,
        filepath: str = None,
        view: bool = True,
        fontsize: int = 10
):
    import matplotlib.pyplot as plt

    features = final_output_test.shape[2]
    if features == 1:
        i = 0
        plt.plot(final_output_test.reshape(-1, final_output_test.shape[-1])[:, i], c='blue', label="measured")
        plt.plot(pred.reshape(-1, pred.shape[-1])[:, i], c='red', label="prediction")
        plt.legend()
        # plt.locator_params(nbins=3)

        plt.xlabel(f'Feature {i + 1} across all timesteps', fontsize=fontsize)
        plt.ylabel(f'F_{i + 1}', fontsize=fontsize)
        # ax.set_title('Title', fontsize=fontsize)
        mse = mean_squared_error(
            final_output_test.reshape(-1, final_output_test.shape[-1])[:, i],
            pred.reshape(-1, pred.shape[-1])[:, i],
            squared=False
        )
        print(f'Feature {i + 1} MSE: {mse}')
        if view:
            plt.show()
        if filepath is not None:
            plt.savefig(filepath)
    else:
        fig, axs = plt.subplots(int(features), 1, layout=None)
        aux = 0
        for ax, i in zip(axs.flat, range(features)):
            _examples_plots(ax, final_output_test, pred, i, fontsize)
            aux += 1
        if view:
            plt.show()
        if filepath is not None:
            fig.savefig(filepath)


def _evaluate_pred_global(
        y_test,
        y_pred,
        model_name=None
):
    list_aux = []
    name_aux = []
    _rmse = tkmtc.RootMeanSquaredError()
    _mae = tkmtc.MeanAbsoluteError()
    _mape = tkmtc.MeanAbsolutePercentageError()
    list_aux.append(float(_mae(y_test, y_pred)))
    name_aux.append('MAE')
    list_aux.append(float(_rmse(y_test, y_pred)))
    name_aux.append('RMSE')
    list_aux.append(float(_mape(y_test, y_pred)))
    name_aux.append('MAPE')
    list_aux.append(model_name)
    name_aux.append('Model')
    df_aux = pd.DataFrame([list_aux], columns=name_aux)
    print(df_aux)

    return df_aux


def _metrics_3D(
        y_colname,
        n_output,
        y_test,
        yhat,
        name_model=None,
        path_save: str = None,
        scaler=None
):
    import pandas as pd
    from sklearn.metrics import mean_squared_error, mean_absolute_error

    if scaler != None:
        d1, d2, d3 = y_test.shape[0], y_test.shape[1], y_test.shape[2]

        y_test = y_test.reshape(d1 * d2, d3)
        yhat = yhat.reshape(d1 * d2, d3)
        y_test = scaler.inverse_transform(y_test)
        yhat = scaler.inverse_transform(yhat)

        y_test = y_test.reshape(d1, d2, d3)
        yhat = yhat.reshape(d1, d2, d3)

    dict_values = {}
    for a in range(len(y_colname)):
        col_true, col_pred, col_mae, col_mse = [], [], [], []
        for i in range(n_output):
            col_true.append(f'True_{i}')
            col_pred.append(f'Pred_{i}')
            col_mae.append(f'MAE_{i}')
            col_mse.append(f'MSE_{i}')
        col_value = col_true + col_pred + col_mae + col_mse
        df_value = pd.DataFrame(
            columns=col_value,
            index=range(y_test.shape[0])
        )

        for jj in range(y_test.shape[0]):
            list_true, list_pred, list_mae, list_mse = [], [], [], []
            for m in range(n_output):
                true, pred = y_test[jj][:, a: a + 1][m], yhat[jj][:, a: a + 1][m]
                list_true.append(true[0])
                list_pred.append(pred[0])
                list_mae.append(mean_absolute_error(true, pred))
                list_mse.append(mean_squared_error(true, pred))
            list_value = list_true + list_pred + list_mae + list_mse
            df_value.iloc[jj] = list_value

        if path_save != None:
            key = y_colname[a]
            df_value.to_csv(
                f'{path_save}/{name_model}_{key}.csv',
                index=False
            )
        dict_values[y_colname[a]] = df_value

    list_mae_vec = []
    col_3d = ['MAE_3D', 'MSE_3D']
    col_2d = []
    for i in range(len(y_colname)):
        col_2d.append(f'MAE_2D_V{i}')
        col_2d.append(f'MSE_2D_V{i}')

    col_3d_2d = col_3d + col_2d
    df_value_3d = pd.DataFrame(
        columns=col_3d_2d,
        index=range(y_test.shape[0])
    )
    for ii in range(y_test.shape[0]):
        list_mae_3d, list_mse_3d = [], []
        list_mae_3d.append(mean_absolute_error(y_test[ii], yhat[ii]))
        list_mse_3d.append(mean_squared_error(y_test[ii], yhat[ii]))

        list_mae_2d, list_mse_2d = [], []
        for a in range(len(y_colname)):
            true, pred = y_test[ii][:, a: a + 1], yhat[ii][:, a: a + 1]
            list_mae_2d.append(mean_absolute_error(true, pred))
            list_mse_2d.append(mean_squared_error(true, pred))
        list_value_3d_2d = list_mae_3d + list_mse_3d + list_mae_2d + list_mse_2d

        df_value_3d.iloc[ii] = list_value_3d_2d

    if path_save != None:
        df_value.to_csv(
            f'{path_save}/{name_model}_3D.csv',
            index=False
        )
    return dict_values, df_value_3d


def _plot_boxplot_n_steps_future(dict_values_2D, df_values_3D):
    import matplotlib.pyplot as plt
    import seaborn as sns
    import re
    metrics = ['MAE', 'MSE']
    for i in metrics:
        filter = [x for x in list(df_values_3D.columns) if re.match(i, x)]
        df_aux = df_values_3D[filter]
        fig, ax = plt.subplots(figsize=(10, 5))
        boxplot = sns.boxplot(data=df_aux.astype(float), showfliers=False)
        # boxplot.axes.set_title("MAE en 3D y 2D", fontsize=16)
        boxplot.set_ylabel(i, fontsize=14)
        # boxplot.set_xlabel("Columnas", fontsize=14)
        plt.show()

    for k in metrics:
        i = 0
        fig, axes = plt.subplots(1, 4, figsize=(12, 3), constrained_layout=True)
        for x in list(dict_values_2D.resolution()):
            filter = [x for x in dict_values_2D[x].columns if re.match(k, x)]
            df_aux = dict_values_2D[x][filter]
            fig = sns.boxplot(data=df_aux, ax=axes[i], showfliers=False)
            fig.axes.set_title(x)
            fig.axes.set_ylabel(k)
            i += 1
        plt.show()


def reshape_test(type_mdl, dict_test):
    if type_mdl == 'MLP':
        X_test = dict_test['X']
        return X_test.reshape(X_test.shape[0], X_test.shape[1] * X_test.shape[2])
    if type_mdl == 'Conv1D':
        X_test = dict_test['X']
        return X_test.reshape(X_test.shape[0], X_test.shape[1], X_test.shape[2], 1)
    if type_mdl == 'Conv2D':
        X_test = dict_test['X']
        return X_test.reshape(X_test.shape[0], X_test.shape[1], X_test.shape[2], 1, 1)
    else:
        return dict_test['X']


def reshape_yhat(type_mdl, yhat, n_future, n_out_ft):
    if type_mdl == 'MLP':
        return yhat.reshape(yhat.shape[0], n_future, n_out_ft)
    else:
        return yhat

class _process_values_preliminary:

    def __init__(
            self,
            model,
            dict_test,
            config_data,
            config_sim,
            name_model=None
    ):
        self.model = model
        self.dict_test = dict_test
        self.config_data = config_data
        self.config_sim = config_sim
        self.name_model = name_model

    def get_values(
            self,
            type_mdl,
            filepath
    ):
        X_test = reshape_test(type_mdl, self.dict_test)
        X_test[np.isnan(X_test)] = value_miss

        yhat = self.model.predict(X_test)
        yhat = reshape_yhat(type_mdl, yhat, self.config_data['n_future'], self.config_data['n_out_ft'])
        yhat_inverse, y_test_inverse = _inverse_transform(yhat, self.dict_test, self.config_data['y_colname'])
        _evaluate_pred_global(y_test_inverse, yhat_inverse, self.name_model)
        if self.config_sim['plt_results']:
            _plot_results(y_test_inverse, yhat_inverse)
            _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, self.config_data)
        if self.config_sim['plt_metrics']:
            dict_values_M1, df_value_3d_M1 = _metrics_3D(
                y_colname=self.config_data['y_colname'],
                n_output=self.config_data['n_future'],
                y_test=y_test_inverse,
                yhat=yhat_inverse
            )
            _plot_boxplot_n_steps_future(dict_values_M1, df_value_3d_M1)
        if self.config_sim['metrics']:
            scorer_dict = dict_metrics(y_test_inverse, yhat_inverse, self.config_data, self.name_model)
            return yhat_inverse, scorer_dict
        else:
            yhat_inverse

        return yhat_inverse


def _values_preliminary(model, dict_test, config_data, config_sim, name_model=None):
    yhat = model.predict(dict_test['X'])
    yhat_inverse, y_test_inverse = _inverse_transform(yhat, dict_test, config_data['y_colname'])
    _evaluate_pred_global(y_test_inverse, yhat_inverse, name_model)
    if config_sim['plt_results']:
        _plot_results(y_test_inverse, yhat_inverse)
        _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config_data)
    if config_sim['plt_metrics']:
        dict_values_M1, df_value_3d_M1 = _metrics_3D(
            y_colname=config_data['y_colname'],
            n_output=config_data['n_future'],
            y_test=y_test_inverse,
            yhat=yhat_inverse
        )
        _plot_boxplot_n_steps_future(dict_values_M1, df_value_3d_M1)
    if config_sim['metrics']:
        scorer_dict = dict_metrics(y_test_inverse, yhat_inverse, config_data, name_model)
        return yhat_inverse, scorer_dict
    else:
        yhat_inverse
    return yhat_inverse


def _values_preliminary_mlp(model, dict_test, config_data, name_model=None):
    X_test = dict_test['X']
    X_test = X_test.reshape(X_test.shape[0], config_data['n_past'] * config_data['n_inp_ft'])
    yhat = model.predict(X_test)
    yhat = yhat.reshape(yhat.shape[0], config_data['n_future'], config_data['n_out_ft'])
    yhat_inverse, y_test_inverse = _inverse_transform(yhat, dict_test, config_data['y_colname'])
    _evaluate_pred_global(y_test_inverse, yhat_inverse, name_model)
    if config_data['plt_results']:
        _plot_results(y_test_inverse, yhat_inverse)
        _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config_data)
    if config_data['plt_metrics']:
        dict_values_M1, df_value_3d_M1 = _metrics_3D(
            y_colname=config_data['y_colname'],
            n_output=config_data['n_future'],
            y_test=y_test_inverse,
            yhat=yhat_inverse
        )
        _plot_boxplot_n_steps_future(dict_values_M1, df_value_3d_M1)
    if config_data['metrics']:
        scorer_dict = dict_metrics(y_test_inverse, yhat_inverse, config_data, name_model)
        return yhat_inverse, scorer_dict
    else:
        yhat_inverse

    return yhat_inverse


def _values_preliminary_1D(model, dict_test, config, name_model=None):
    X_test = dict_test['X']
    X_test = X_test.reshape(X_test.shape[0], X_test.shape[1], X_test.shape[2], 1)
    yhat = model.predict(X_test)
    yhat = yhat.reshape(yhat.shape[0], yhat.shape[1], yhat.shape[2])
    yhat_inverse, y_test_inverse = _inverse_transform(yhat, dict_test, config['y_colname'])
    _evaluate_pred_global(y_test_inverse, yhat_inverse, name_model)
    if config['plt_results']:
        _plot_results(y_test_inverse, yhat_inverse)
        _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config)
    if config['plt_metrics']:
        dict_values_M1, df_value_3d_M1 = _metrics_3D(
            y_colname=config['y_colname'],
            n_output=config['n_future'],
            y_test=y_test_inverse,
            yhat=yhat_inverse
        )
        _plot_boxplot_n_steps_future(dict_values_M1, df_value_3d_M1)
    if config['metrics']:
        scorer_dict = dict_metrics(y_test_inverse, yhat_inverse, config, name_model)
        return yhat_inverse, scorer_dict
    else:
        yhat_inverse

    return yhat_inverse


def _values_preliminary_2D(model, dict_test, config, name_model=None):
    X_test = dict_test['X']
    X_test = X_test.reshape(X_test.shape[0], X_test.shape[1], X_test.shape[2], 1, 1)

    yhat = model.predict(X_test)
    yhat = yhat.reshape(yhat.shape[0], yhat.shape[1], yhat.shape[2])

    yhat_inverse, y_test_inverse = _inverse_transform(yhat, dict_test, config['y_colname'])
    _evaluate_pred_global(y_test_inverse, yhat_inverse, name_model)
    if config['plt_results']:
        _plot_results(y_test_inverse, yhat_inverse)
        _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config)
    if config['plt_metrics']:
        dict_values_M1, df_value_3d_M1 = _metrics_3D(
            y_colname=config['y_colname'],
            n_output=config['n_future'],
            y_test=y_test_inverse,
            yhat=yhat_inverse
        )
        _plot_boxplot_n_steps_future(dict_values_M1, df_value_3d_M1)
    if config['metrics']:
        scorer_dict = dict_metrics(y_test_inverse, yhat_inverse, config, name_model)
        return yhat_inverse, scorer_dict
    else:
        yhat_inverse

    return yhat_inverse


def _values_preliminary_mdn(model, dict_test, config, output_dim, n_mixtures, name_model=None):
    from mdn import sample_from_output

    yhat = model.predict(dict_test['X'])
    pred_dist = []
    for i in range(0, 10001):
        y_samples = np.apply_along_axis(sample_from_output, 1, yhat, output_dim, n_mixtures, temp=1.0)
        pred_dist.append(y_samples[0].ravel())

    yhat = pred_dist.reshape(pred_dist.shape[0], config['n_future'], config['n_out_ft'])
    yhat_inverse, y_test_inverse = _inverse_transform(yhat, dict_test, config['y_colname'])
    _evaluate_pred_global(y_test_inverse, yhat_inverse, name_model)
    if config['plt_results']:
        _plot_results(y_test_inverse, yhat_inverse)
        _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config)
    if config['plt_metrics']:
        dict_values_M1, df_value_3d_M1 = _metrics_3D(
            y_colname=config['y_colname'],
            n_output=config['n_future'],
            y_test=y_test_inverse,
            yhat=yhat_inverse
        )
        _plot_boxplot_n_steps_future(dict_values_M1, df_value_3d_M1)
    if config['metrics']:
        scorer_dict = dict_metrics(y_test_inverse, yhat_inverse, config, name_model)
        return yhat_inverse, scorer_dict
    else:
        yhat_inverse

    return yhat_inverse


def _values_rdr_score_mlp(model, dict_train, dict_test, config):
    # train
    X_train = dict_train['train']['X']

    # test
    X_test = dict_test['X']
    X_test = X_test.reshape(X_test.shape[0], config['n_past'] * config['n_inp_ft'])

    yhat = model.predict(X_test)
    yhat = yhat.reshape(yhat.shape[0], config['n_future'], config['n_out_ft'])
    yhat_inverse, y_test_inverse = _inverse_transform(yhat, dict_test, config['y_colname'])
    # test_rdr_scorer(X_train, yhat_inverse, y_test_inverse, config)


def dict_metrics(y_test_inverse, yhat_inverse, config_data, name_model):
    ytest_2d = y_test_inverse.reshape(y_test_inverse.shape[0] * y_test_inverse.shape[1], y_test_inverse.shape[2])
    yhat_2d = yhat_inverse.reshape(yhat_inverse.shape[0] * yhat_inverse.shape[1], yhat_inverse.shape[2])
    df_ytest = pd.DataFrame(ytest_2d, columns=config_data['y_colname'])
    df_yhat = pd.DataFrame(yhat_2d, columns=config_data['y_colname'])
    scorer_dict = {}
    for idx, y_colname in enumerate(config_data['y_colname']):
        scorer = calculate_metrics()
        scorer.fit(
            y_true=df_ytest[y_colname],
            y_pred=df_yhat[y_colname],
            model_name=y_colname + '_' + name_model
        )
        scorer_dict[y_colname] = scorer
    return scorer_dict


def _R_squared(y_true, y_pred):
    """
    R^2 (coefficient of determination) regression score function.
    Best possible score is 1.0, lower values are worse.
    Args:
        y_true ([np.array]): test samples
        y_pred ([np.array]): predicted samples
    Returns:
        [float]: R2
    """
    SS_res = tf.reduce_sum(tf.square(y_true - y_pred), axis=-1)
    SS_tot = tf.reduce_sum(tf.square(y_true - tf.reduce_mean(y_true, axis=-1)), axis=-1)
    return (1 - SS_res / (SS_tot + tf.keras.backend.epsilon()))


def _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config, view=True, filepath=None):
    import matplotlib.pyplot as plt
    _rmse = tkmtc.RootMeanSquaredError()
    _mape = tkmtc.MeanAbsolutePercentageError()
    _mae = tkmtc.MeanAbsoluteError()

    # yhat_inverse_time_step and y_test_inverse_time_step are both same dimension.
    time_step_list_yhat = [[] for i in range(config['n_future'])]
    time_step_list_y_test = [[] for i in range(config['n_future'])]
    for i in range(0, yhat_inverse.shape[0]):
        for j in range(0, yhat_inverse.shape[1]):
            time_step_list_yhat[j].append(list(yhat_inverse[i][j]))
            time_step_list_y_test[j].append(list(y_test_inverse[i][j]))
    yhat_time_step = np.array(time_step_list_yhat)
    yhat_time_step = yhat_time_step.reshape(yhat_time_step.shape[0], -1)
    y_test_time_step = np.array(time_step_list_y_test)
    y_test_time_step = y_test_time_step.reshape(y_test_time_step.shape[0], -1)
    # plotting
    mae_list, r2_list, rmse_list = [], [], []
    for i in range(0, config['n_future']):
        mae = _mae(y_test_time_step[i], yhat_time_step[i])
        r2 = _R_squared(y_test_time_step[i], yhat_time_step[i])
        rmse = _rmse(y_test_time_step[i], yhat_time_step[i])
        mae_list.append(mae)
        r2_list.append(r2)
        rmse_list.append(rmse)

    list_name = ['MAE', 'R2', 'RMSE']
    list_results = [mae_list, r2_list, rmse_list]

    dict_results = dict(zip(list_name, list_results))
    fig, axs = plt.subplots(3, 1, layout=None)
    aux = 0
    for key, values in dict_results.items():
        axs[aux].plot(range(0, config['n_future']), values, marker='o')
        axs[aux].set(ylabel=key, xlabel='number of steps in future')
        aux += 1
    if view:
        plt.show()
    if filepath is not None:
        fig.savefig(filepath)



import pandas as pd
from scipy import stats
import math


class MultiplesArchivos(object):

    dfList = []  # Declara una lista vacía
    dfList_User = []
    Ener = pd.DataFrame()
    df_Season = pd.DataFrame()
    df_Season_Stadistic = pd.DataFrame()
    noFile = int
    progress = int

    def __init__(self, indir, temp, t_day):
        self.temp = temp
        self.indir = indir
        self.t_day = t_day

    def user(self, df):
        """
        Método para leer atributos de los usuraios y el ID
        """
        global Terminal
        Terminal = df['Unnamed: 1'][1]  # obtien valores de la columna 'Unnamed: 1' inice 1
        # Domicilio = df['Unnamed: 1'][2]
        # Servicio = df['Unnamed: 1'][3]
        # Periodo = df['Unnamed: 8'][2]
        # datos = pd.DataFrame({'Terminal': [Terminal],'Domicilio':[Domicilio],'Servicio':[Servicio]})
        return Terminal

    def cleanExcel(self, df):
        """
        Metodo que permite realizar la limpieza de la hoja de datos (panilla excel)
        """
        # axis es 0 para fila, 1 para columna
        df = df.drop(
            labels=[
                'Unnamed: 1', 'Unnamed: 2', 'Unnamed: 3', 'Unnamed: 5', 'Unnamed: 6',
                'Unnamed: 7', 'Unnamed: 8', 'Unnamed: 9', 'Unnamed: 10', 'Unnamed: 12'
            ],
            axis=1
            )  # elimana la columna 'Unnamed: 1' ......
        df = df.drop([0, 1, 2, 3, 4, 5, 6, 7])  # Eliminar las filas.
        df = df.reset_index(drop=True)  # Reiniciar al indexado.
        # Cambiar los nombres de las columnas
        df = df.rename(index=str, columns={
            "Reporte de Registros Periódicos": "Fecha hora",
            # "Unnamed: 2": "Tensión(V)",
            # "Unnamed: 3": "Corriente",
            "Unnamed: 4": "E_Activa(Wh)",
            # "Unnamed: 5": "E_Aparente(VARh)",
            # "Unnamed: 7": "E_Reactiva(VAh)",
            # "Unnamed: 9": "Frec(Hz)",
            # "Unnamed: 10": "FP",
            "Unnamed: 11": "D.Max(W)"}
        )
        df = df.dropna(how='any')
        return df

    def proces_DF(self, df):
        """
        Códigos para obtener por separdos en diferentes columnas del dataframe /n horas menses y días de la semana /n
        """
        global Min
        global Horas
        global Dias_S
        global Dia_M
        global Mes
        global year

        df['Fecha hora'] = pd.to_datetime(df['Fecha hora'], errors='coerce')
        Horas = df['Fecha hora'].dt.hour        # con la propiedad .dt podemos acceder a los difrentes intervalos de fecha y hora.
        Dias_S = df['Fecha hora'].dt.weekday    # Obtener los días de la semana, cero para lunes, y 6 para domingo.
        Dia_M = df['Fecha hora'].dt.day
        Mes = df['Fecha hora'].dt.month         # obtener los meses de la columna fecha.
        Min = df['Fecha hora'].dt.minute        # obtener los minutos.
        year = df['Fecha hora'].dt.year
        pot_media = df['E_Activa(Wh)'] / 0.25   # Se calcula la pontencia activa instantánea\n",

        # Añadimos la columna al dataframe de las potencias unas vez calculadas.
        user_data_temp = df.copy()
        user_data_temp['Pot_media(W)'] = pot_media  # Añadimos la columna Potencia activa
        user_data_temp['Tipo de Día'] = Dias_S      # Añadimos la columna días de la semana.
        user_data_temp['Dia del Mes'] = Dia_M
        user_data_temp['Hora del Día'] = Horas
        user_data_temp['Mes'] = Mes
        user_data_temp['Año'] = year
        df = user_data_temp
        return df

    def fil_type_day(self, df):
        """
        Metodo que filtra por tipo de día de la semana seleccionado
        """
        if self.t_day == 'habil':
            df_type_day = df[(df['Tipo de Día'] <= 4)]  # Seleccionamos los días hábileses un dataframe donde solo se incluyen los días Hábiles.
        elif self.t_day == 'no habil':
            df_type_day = df[(df['Tipo de Día'] >= 5)]  # selección de los días no hábiles.
        return df_type_day

    def fil_season(self, df):
        """
        Metodo que filtra el dataframe que recibe por la estación del año selecionada
        """
        # global df_season
        verano = [12, 1, 2]
        invierno = [6, 7, 8]
        intermedio = [3, 4, 5, 9, 10, 11]
        if self.temp == 'verano':
            df_type_season = df[df['Mes'].isin(verano)]
        elif self.temp == 'invierno':
            df_type_season = df[df['Mes'].isin(invierno)]
        elif self.temp == 'otoño y primavera':
            df_type_season = df[df['Mes'].isin(intermedio)]
        return df_type_season

    def maxDeman(self, df, user):
        """
        Método que recibe el dataframe y la lista del usuario para devolver un dataframe compuesta por energías y
        potencias indexadas por el codigo del usuario
        """
        '''        
        Df_Temp = pd.DataFrame()
        Df_Temp['D.Max(W)'] = df['D.Max(W)'].max()
        Df_Temp['Usuario'] = user

        user_data_temp = Df_Temp.copy()
        user_data_temp['Usuario'] = user  # se añade la columna de los codigos a la el dataframe de los datos
        result = user_data_temp.set_index('Usuario')  # se coloca codigos de los usuarios como indices de la tabla promedio
        '''
        result = pd.DataFrame([[df['D.Max(W)'].max(), user]], columns=['D.Max(W)', 'Usuario'])
        return result

    def ener_hora_acum(self, df):
        """
        Este método devuelve un DataFrame que contiene columnas con los valores de las energías horarias acumuladas
        durante las 24 horas las filas son las mediciones totales agrupadas cada 15 minutos esto representa la cantidad
        de dias, siendo este acumulado perteneciente al número de día en que se obtuvo medicion.
        """
        Listdf_clean = pd.DataFrame()  # datasets frame vacío.
        for h in range(0, 24):  # h va a ieterar 24 veces representando las 24 horas del día.
            df_clean = pd.DataFrame()
            k = [h]
            Hora_Dia = df[df['Hora del Día'].isin(k)]  # filtramos por hora del dia
            Hora_Dia_Acumu = Hora_Dia.groupby(
                [year, Mes, Dia_M, Horas])['E_Activa(Wh)'].sum()  # Sumar las mediciones de los periodos de 15 min entre horas
            Hora_Dia_Acum_ = pd.DataFrame(pd.DataFrame(Hora_Dia_Acumu))  # convertir a datasets frame a dataframe
            Hora_Dia_Acum_ = Hora_Dia_Acum_.reset_index(drop=True)  # se elimina niveles de indices

            vector = Hora_Dia_Acum_['E_Activa(Wh)']
            name_colum = '$%d$' % h  # nombre de la columna
            df_clean = df_clean.copy()

            df_clean[name_colum] = vector  # coloca el vector hora h en la columna h de df_celan
            Listdf_clean = pd.concat([Listdf_clean, df_clean], axis=1)  # adjuntos el nuevo vector de las horas a al dataframe
            Listdf_clean = Listdf_clean.dropna(how='any')  # Elimino datos null de df
            Listdf_clean.columns = range(Listdf_clean.shape[1])  # cambiamos el tipo de datos de los nombres de las columnas para dierecccionar luego

        return Listdf_clean

    def prom_clean(self, df, User):
        """"
        Este método elimina los datos atípicos y entrega un vector de la energía promedio horario indexado por ID de usuario
        """
        z_critical = stats.norm.ppf(q=0.999)  # Get the z-critical value* determina el
        data_graf = []
        vector_Prom = []
        num = len(df.columns)
        dfData_graf = pd.DataFrame()
        if df.empty:
            print('here')

        for i in range(num):
            data = df[i]
            sample_size = len(data)     # determinar la longitud de los datos
            if sample_size != 0:
                sample_mean = data.mean()   # determina la media.
                pop_stdev = data.std()      # Get the population standard deviation
                margin_of_error = z_critical * (pop_stdev / math.sqrt(sample_size))  # margen de error
                confidence_interval = (sample_mean - margin_of_error, sample_mean + margin_of_error)
                interv_min = sample_mean - margin_of_error
                interv_max = sample_mean + margin_of_error
                data = data[(data >= interv_min)]  # filtrar obteniendo los valores de mayores al min
                data = data[(data <= interv_max)]  # filtrar obteniendo los valores de menores al min
            data_graf.append(data)
            # P = pd.DataFrame(datasets)
            # dfData_graf =  dfData_graf.join(P)
            Promedio = data_graf[i].mean()
            vector_Prom.append(Promedio)

        # Normalizar datos
        vector_Prom = pd.DataFrame(vector_Prom)
        # scaler = MinMaxScaler(feature_range=(0,1))
        # vector_Prom = scaler.fit_transform(vector_Prom)
        # vector_Prom = normalice(vector_Prom)
        vector_Prom_T = pd.DataFrame(vector_Prom).T

        vector = vector_Prom_T.copy()
        vector['Usuario'] = User  # se añade la columna de los codigos a la el dataframe de los datos
        # print(user_data_temp)
        vector = vector.set_index('Usuario')  # se coloca codigos de los usuarios como indices de la tabla promedio
        # vector.fillna(method='ffill', axis=1, inplace = True) # Se rellenan los campos nulos que pudieran existir
        return vector

    def stadistic(self, df1, df2):
        df_temp = pd.DataFrame()
        # media = df1.median(axis=1)
        desv_estandar = df1.std(axis=1)
        # kurtosis  = df1.kurtosis(axis=1)  # compara cantidad de datos cerca de la media con lo que esat lejos
        # asimetria  = df1.skew(axis=1)     # calculo de asimetría de los datos sobre la media
        Ener_Pro_Dia = df1.sum(axis=1)
        Deman_Max = df2['D.Max(W)'].max()

        # df_temp['Media'] = media
        df_temp['dev_Estandar'] = desv_estandar
        # df_temp['kurtosis'] = kurtosis
        df_temp['Ener_Pro_Dia'] = Ener_Pro_Dia
        # df_temp['Asimetría'] = asimetria
        df_temp['Dema_Pico'] = Deman_Max
        return df_temp



from openpy_fxts.base_tf import tkm, tkl
from openpy_fxts.architectures.layers.layers_class import multi_conv1d_pool_flat_drop, dense_multi_out
from openpy_fxts.architectures.fxts.arch_dicts import _get_dict_fxts


class conv1D_dense(tkm.Model):

    def __init__(
            self,
            n_future: int = None,
            n_out_ft: int = None,
            config: dict = None
    ):
        super(conv1D_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('Conv1D_Dense').config_conv1d()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.multi_conv1d = multi_conv1d_pool_flat_drop(
            filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            activations=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            maxpooling_1d=True,
            pool_size=config['conv1d_layers']['pool_size']
        )
        if config['dense_layer']['activation']:
            self.dense_layer = tkl.Dense(
                units=config['dense_layer']['units'], 
                activation=config['dense_layer']['activation']
            )
        if config['dropout']['activate']:
            self.drop = tkl.Dropout(rate=config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_conv1d(inputs)
        if self.config['dense_layer']['activation']:
            x = self.dense_layer(x)
            if self.config['dropout']['activate']:
                x = self.drop(x)
        x = self.dense_out(x)
        return x


class multi_conv1D_dense(tkm.Model):

    def __init__(
            self,
            n_inp_ft: int = None,
            n_future: int = None,
            n_out_ft: int = None,
            config: dict = None
    ):
        super(multi_conv1D_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('Multi_Conv1D_Dense').config_conv1d()
        self.n_inp_ft = n_inp_ft
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        filters = config['conv1d_layers']['filters']
        kernels = config['conv1d_layers']['kernels']
        activations = config['conv1d_layers']['activations']
        paddings = config['conv1d_layers']['paddings']
        self.hidden_layers = []
        for i in range(n_inp_ft):
            self.hidden_layers.append(
                multi_conv1d_pool_flat_drop(
                    filters=filters,
                    kernels=kernels,
                    activations=activations,
                    paddings=paddings,
                    use_drop=config['dropout']['activate'],
                    rate=config['dropout']['rate'],
                    maxpooling_1d=True,
                    pool_size=config['conv1d_layers']['pool_size']
                )
            )
        self.concat = tkl.Concatenate(axis=1)
        if config['dense_layer']['activation']:
            self.dense_layer = tkl.Dense(
                units=config['dense_layer']['units'],
                activation=config['dense_layer']['activation']
            )
            self.drop = tkl.Dropout(rate=config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        head_list = []
        for i, (layer_conv1d) in enumerate(self.hidden_layers):
            head = layer_conv1d(inputs)
        head_list.append(head)
        x = self.concat(head_list)
        if self.config['dense_layer']['activation']:
            x = self.dense_layer(x)
            x = self.drop(x)
        x = self.dense_out(x)
        return x


# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.architectures.baseline_arch import base_arch
from openpy_fxts.base_tf import tkl, tkm, value_miss
# Simple
# models_fxts -> MLP
from openpy_fxts.architectures.mdits.simple.MLP_class import mlp_dense_class

# models_fxts -> BiRNN
# Simple
from openpy_fxts.architectures.mdits.simple.RNN_class import birnn_dense_class, multi_birnn_dense_class
# Hybrids
from openpy_fxts.architectures.mdits.hybrids.BiRNN_others import birnn_conv1d_dense, birnn_timedist_dense_class, \
    birnn_bahdanau_attention_conv1d_dense, birnn_luong_attention_conv1d_dense, birnn_multihead_attention_conv1d_dense
# Models -> Conv1D
# Simple
from openpy_fxts.architectures.mdits.simple.Conv1D_class import conv1D_dense, multi_conv1D_dense
# Hybrids
from openpy_fxts.architectures.mdits.hybrids.Conv1D_others import conv1d_birnn_dense
from openpy_fxts.architectures.mdits.hybrids.Conv1D_others import timedist_conv1d_birnn_dense
from openpy_fxts.architectures.mdits.hybrids.Conv1D_others import conv1d_birnn_attention_dense
# models_fxts -> Others
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Others import tcn_bilstm
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Others import time2vec_bilstm
# models_fxts -> EncDec
from openpy_fxts.architectures.mdits.hybrids.EncDec_hybrids import encdec_birnn, encdec_conv1d_birnn


# models_fxts -> LSTM
from openpy_fxts.architectures.mdits.hybrids.class_mdls.LSTM_others import lstm2_dense
# models_fxts -> BiLSTM
from openpy_fxts.architectures.mdits.hybrids.class_mdls.BiLSTM_others import bilstm_dense
from openpy_fxts.architectures.mdits.hybrids.class_mdls.BiLSTM_others import bilstm_conv1d_dense
from openpy_fxts.architectures.mdits.hybrids.class_mdls.BiLSTM_others import bilstm_bahdanau_attention_conv1d_dense
from openpy_fxts.architectures.mdits.hybrids.class_mdls.BiLSTM_others import bilstm_multihead_attention_conv1d_dense
from openpy_fxts.architectures.mdits.hybrids.class_mdls.BiLSTM_others import bilstm_luong_attention_conv1d_dense
# models_fxts -> Conv1D
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Conv1D_others import conv1D_lstm_dense
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Conv1D_others import conv1D_bilstm_dense
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Conv1D_others import conv1D_bilstm_attention_dense
# models_fxts -> GRU
from openpy_fxts.architectures.mdits.hybrids.class_mdls.GRU_others import gru2_dense
# models_fxts -> Stacked
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Stacked_others import lstm_stacked
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Stacked_others import bilstm_stacked
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Stacked_others import stackedscinet
# models_fxts -> Seq2Seq
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_lstm
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_lstm2
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_lstm_batch_drop
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_bilstm
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_bilstm2
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_conv1d_bilstm
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_multihead_conv1d_bilstm  # Function
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_bilstm_with_attention  # Function
from openpy_fxts.architectures.mdits.hybrids.class_mdls.Seq2Seq_hybrids import seq2seq_lstm_with_loung_attention  # Function


class mdits_mdls(base_arch):

    def __init__(self, name_mdl, type_mdl, n_past, n_future, n_inp_ft, n_out_ft, config_arch):
        super().__init__(name_mdl, type_mdl, n_past, n_future, n_inp_ft, n_out_ft, config_arch)

    def models(self):
        if self.type_mdl == 'MLP':
            return self._mlp_mdls()
        if self.type_mdl == 'Conv1D':
            return self._conv1d_mdls()
        if self.type_mdl == 'BiRNN':
            return self._birnn_mdls()
        if self.type_mdl == 'Others':
            return self._others_mdls()
        if self.type_mdl == 'EncDec':
            return self._encdec_mdls()
        if self.type_mdl == 'LSTM':
            return self._lstm_mdls()
        if self.type_mdl == 'GRU':
            return self._gru_mdls()
        if self.type_mdl == 'BiLSTM':
            return self._bilstm_mdls()
        if self.type_mdl == 'Stacked':
            return self._stacked_mdls()
        if self.type_mdl == 'Seq2Seq':
            return self._seq2seq_mdls()

    def _mlp_mdls(self):
        input_layer = []
        for i in range(self.n_inp_ft):
            input_layer.append(tkl.Input(shape=(self.n_past,)))
        if self.name_mdl == 'MLP_Dense':
            x = mlp_dense_class(self.n_past, self.n_inp_ft, self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)

    def _birnn_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        input_layer = tkl.Masking(mask_value=value_miss)(input_layer)
        # Simple
        if self.name_mdl == 'BiRNN_Dense':
            x = birnn_dense_class(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'Multi_BiRNN_Dense':
            x = multi_birnn_dense_class(
                self.n_past, self.n_inp_ft, self.n_future, self.n_out_ft, self.config_arch
            )(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        # Hybrids
        if self.name_mdl == 'BiRNN_Conv1D':
            x = birnn_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=[input_layer], outputs=x, name=self.name_mdl)
        if self.name_mdl == 'BiRNN_TimeDist_Dense':
            x = birnn_timedist_dense_class(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'BiRNN_Bahdanau_Attention_Conv1D':
            x = birnn_bahdanau_attention_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'BiRNN_Luong_Attention_Conv1D':
            x = birnn_luong_attention_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'BiRNN_MultiHeadAttention_Conv1D':
            x = birnn_multihead_attention_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)

    def _others_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        input_layer = tkl.Masking(mask_value=value_miss)(input_layer)
        if self.name_mdl == 'TCN_BiRNN':
            x = tcn_bilstm(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Time2Vec_BiRNN':
            x = time2vec_bilstm(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)

    def _encdec_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        if self.name_mdl == 'EncDec_BiRNN':
            x = encdec_birnn(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'EncDec_Conv1D_BiRNN':
            x = encdec_conv1d_birnn(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)

    def _lstm_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        if self.name_mdl == 'LSTM2_Dense':
            x = lstm2_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)

    def _gru_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        if self.name_mdl == 'GRU2_Dense':
            x = gru2_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)

    def _bilstm_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        if self.name_mdl == 'BiLSTM_Dense':
            x = bilstm_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=[input_layer], outputs=x)
        if self.name_mdl == 'BiLSTM_Conv1D':
            x = bilstm_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=[input_layer], outputs=x)
        if self.name_mdl == 'BiLSTM_Bahdanau_Attention_Conv1D':
            x = bilstm_bahdanau_attention_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'BiLSTM_MultiHeadAttention_Conv1D':
            x = bilstm_multihead_attention_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'BiLSTM_Luong_Attention_Conv1D':
            x = bilstm_luong_attention_conv1d_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)

    def _conv1d_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        # Simple
        if self.name_mdl == 'Conv1D_Dense':
            x = conv1D_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'Multi_Conv1D_Dense':
            x = multi_conv1D_dense(
                self.n_inp_ft, self.n_future, self.n_out_ft, self.config_arch
            )(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        # Hybrids
        if self.name_mdl == 'Conv1D_BiRNN':
            x = conv1d_birnn_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'TimeDist_Conv1D_BiRNN':
            input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft, 1))
            input_layer = tkl.Masking(mask_value=value_miss)(input_layer)
            x = timedist_conv1d_birnn_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x, name=self.name_mdl)
        if self.name_mdl == 'Conv1D_BiRNN_Attention':
            x = conv1d_birnn_attention_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=[input_layer], outputs=x, name=self.name_mdl)


        if self.name_mdl == 'Conv1D_LSTM':
            x = conv1D_lstm_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Conv1D_BiLSTM':
            x = conv1D_bilstm_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Conv1D_BiLSTM_Attention':
            x = conv1D_bilstm_attention_dense(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=[input_layer], outputs=x)

    def _stacked_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        if self.name_mdl == 'LSTM_Stacked':
            x = lstm_stacked(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'BiLSTM_Stacked':
            x = bilstm_stacked(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Stacked_SciNet':
            x = stackedscinet(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)

    def _seq2seq_mdls(self):
        input_layer = tkl.Input(shape=(self.n_past, self.n_inp_ft))
        if self.name_mdl == 'Seq2Seq_LSTM':
            x = seq2seq_lstm(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Seq2Seq_LSTM2':
            x = seq2seq_lstm2(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Seq2Seq_LSTM_Batch_Drop':
            x = seq2seq_lstm_batch_drop(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Seq2Seq_BiLSTM':
            x = seq2seq_bilstm(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Seq2Seq_BiLSTM2':
            x = seq2seq_bilstm2(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Seq2Seq_Conv1D_BiLSTM':
            x = seq2seq_conv1d_bilstm(self.n_future, self.n_out_ft, self.config_arch)(input_layer)
            return tkm.Model(inputs=input_layer, outputs=x)
        if self.name_mdl == 'Seq2Seq_Multi_Head_Conv1D_BiLSTM':
            return seq2seq_multihead_conv1d_bilstm(self.n_past, self.n_inp_ft, self.n_future, self.n_out_ft)
        if self.name_mdl == 'Seq2Seq_BiLSTM_with_Attention':
            return seq2seq_bilstm_with_attention(self.n_past, self.n_inp_ft, self.n_future, self.n_out_ft)
        if self.name_mdl == 'Seq2Seq_LSTM_with_Luong_Attention':
            return seq2seq_lstm_with_loung_attention(self.n_past, self.n_inp_ft, self.n_future, self.n_out_ft)



# -*- coding: utf-8 -*-
# @Time    : 25/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
import numpy as np
import miceforest as mf
from sklearn.impute import KNNImputer
from xgbimputer import XGBImputer
from pyppca import ppca
from fancyimpute import SoftImpute, BiScaler
from openpy_fxts.models.imputation.base_lib import init_models

# import sys
# sys.modules['sklearn.neighbors.base'] = sklearn.neighbors._base
# from missingpy import MissForest

from openpy_fxts.models.imputation.utilities.Utils import Preprocessing, Metrics, Utils

prepro = Preprocessing()
metrics = Metrics()


class imp_ml(init_models):
    
    def __init__(
            self,
            df_miss: pd.DataFrame = None,
            df_true: pd.DataFrame = None
    ):
        super().__init__(df_miss, df_true)

    def miceforest(
            self, 
            n_iter: int = 5, 
            seed: int = 123
    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        self.df_miss, transf = prepro.LabelEncoder(self.df_miss)
        kernel = mf.ImputationKernel(
            data=self.df_miss,
            save_all_iterations=True,
            random_state=seed)
        kernel.mice(n_iter, verbose=True)
        MiceForest = kernel.complete_data(dataset=0, inplace=False)
        MiceForest['datetime'] = self.df_miss.index
        MiceForest.set_index('datetime', inplace=True)
        MiceForest = prepro.inverse_LabelEncoder(MiceForest, transf)

        if self.df_true is None:
            return MiceForest
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(self.df_true[list_aux], MiceForest[list_aux], dict_aux, 'mice_forest')
            return MiceForest, dict_aux['mice_forest']

    def knn(
            self,
            missing_values=np.nan,
            n_neighbors=5,
            weights="uniform",
            metric="nan_euclidean",
            copy=True,
            add_indicator=False,
            keep_empty_features=False

    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        self.df_miss, transf = prepro.LabelEncoder(self.df_miss)
        model_knn = KNNImputer(
            missing_values=missing_values,
            n_neighbors=n_neighbors,
            weights=weights,
            metric=metric,
            copy=copy,
            add_indicator=add_indicator,
            keep_empty_features=keep_empty_features
        )  # KNN mdits
        knn_imp = model_knn.fit_transform(self.df_miss)
        knn_imp = pd.DataFrame(knn_imp, columns=self.df_miss.columns).round(1)
        knn_imp['datetime'] = self.df_miss.index
        knn_imp.set_index('datetime', inplace=True)

        if self.df_true is None:
            return knn_imp
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(self.df_true[list_aux], knn_imp[list_aux], dict_aux, 'knn')
            return knn_imp, dict_aux['knn']

    def XGBImputer(
            self, 
    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        model_XGBimp = XGBImputer(
            categorical_features_index=prepro.index_column(self.df_miss),
            replace_categorical_values_back=True
        )
        XGB_Imp = model_XGBimp.fit_transform(self.df_miss)
        XGB_Imp = pd.DataFrame(XGB_Imp, columns=self.df_miss.columns)
        XGB_Imp['datetime'] = self.df_miss.index
        XGB_Imp.set_index('datetime', inplace=True)

        if self.df_true is None:
            return XGB_Imp
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(
                self.df_true[list_aux],
                XGB_Imp[list_aux],
                dict_aux,
                'XGB'
            )
            return XGB_Imp, dict_aux['XGB']

    def ppca(
            self, 
            d=50,
            dia=True
    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        self.df_miss, transf = prepro.LabelEncoder(self.df_miss)
        X_miss = self.df_miss.to_numpy()
        C1, ss1, M1, X1, Ye1 = ppca(X_miss, d, dia)
        ppca_imp = pd.DataFrame(Ye1, columns=self.df_miss.columns)
        ppca_imp['datetime'] = self.df_miss.index
        ppca_imp.set_index('datetime', inplace=True)

        if self.df_true is None:
            return ppca_imp
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(self.df_true[list_aux], ppca_imp[list_aux], dict_aux, 'PPCA')
            return ppca_imp, dict_aux['PPCA']

    def biscaler(
            self,
            shrinkage_value=None,
            convergence_threshold=0.001,
            max_iters=100,
            max_rank=None,
            n_power_iterations=1,
            init_fill_method="zero",
            min_value=None,
            max_value=None,
            normalizer=None,
            verbose=True,
            center_rows=True,
            center_columns=True,
            scale_rows=True,
            scale_columns=True,
            tolerance=0.001,
    ):

        list_aux = Utils(self.df_miss).list_column_missing()
        self.df_miss, transf = prepro.LabelEncoder(self.df_miss)

        # Instead of solving the nuclear norm objective directly, instead
        # induce sparsity using singular value thresholding
        softImpute = SoftImpute(
            shrinkage_value,
            convergence_threshold,
            max_iters,
            max_rank,
            n_power_iterations,
            init_fill_method,
            min_value,
            max_value,
            normalizer,
            verbose
        )

        # simultaneously normalizes the rows and columns of your observed datasets,
        # sometimes useful for low-rank mdits methods
        model_biscaler = BiScaler(
            center_rows,
            center_columns,
            scale_rows,
            scale_columns,
            min_value,
            max_value,
            max_iters,
            tolerance,
            verbose
        )
        X_miss = self.df_miss.to_numpy()
        X_miss_norm = model_biscaler.fit_transform(X_miss)
        X_fill_soft_norm = softImpute.fit_transform(X_miss_norm)
        biscaler_imp = model_biscaler.inverse_transform(X_fill_soft_norm)
        biscaler_imp = pd.DataFrame(biscaler_imp, columns=self.df_miss.columns)
        biscaler_imp['datetime'] = self.df_miss.index
        biscaler_imp.set_index('datetime', inplace=True)

        if self.df_true is None:
            return biscaler_imp
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(
                self.df_true[list_aux],
                biscaler_imp[list_aux],
                dict_aux, 'model_biscaler'
            )
            return biscaler_imp, dict_aux['model_biscaler']

    def softimpute(
            self,
            shrinkage_value=None,
            convergence_threshold=0.001,
            max_iters=100,
            max_rank=None,
            n_power_iterations=1,
            init_fill_method="zero",
            min_value=None,
            max_value=None,
            normalizer=None,
            verbose=True
    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        self.df_miss, transf = prepro.LabelEncoder(self.df_miss)

        # Instead of solving the nuclear norm objective directly, instead
        # induce sparsity using singular value thresholding
        model_softImpute = SoftImpute(
            shrinkage_value,
            convergence_threshold,
            max_iters,
            max_rank,
            n_power_iterations,
            init_fill_method,
            min_value,
            max_value,
            normalizer,
            verbose
        )
        X_miss = self.df_miss.to_numpy()
        soft_imp = model_softImpute.fit_transform(X_miss)
        soft_imp = pd.DataFrame(soft_imp, columns=self.df_miss.columns)
        soft_imp['datetime'] = self.df_miss.index
        soft_imp.set_index('datetime', inplace=True)

        if self.df_true is None:
            return soft_imp
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(self.df_true[list_aux], soft_imp[list_aux], dict_aux, 'softimpute')
            return soft_imp, dict_aux['softimpute']



from openpy_fxts.mdits.mdls_dl.base_tf import tkl, K


class Time2Vec(tkl.Layer):

    def __init__(self, output_dim=None, **kwargs):
        self.output_dim = output_dim
        super(Time2Vec, self).__init__(**kwargs)

    def build(self, input_shape):
        self.W = self.add_weight(
            name='W',
            shape=(input_shape[-1], self.output_dim),
            initializer='uniform',
            trainable=True)

        self.P = self.add_weight(
            name='P',
            shape=(input_shape[1], self.output_dim),
            initializer='uniform',
            trainable=True
        )
        self.w = self.add_weight(
            name='w',
            shape=(input_shape[1], 1),
            initializer='uniform',
            trainable=True
        )

        self.p = self.add_weight(
            name='p',
            shape=(input_shape[1], 1),
            initializer='uniform',
            trainable=True
        )
        super(Time2Vec, self).build(input_shape)

    def call(self, x):
        original = self.w * x + self.p
        sin_trans = K.sin(K.dot(x, self.W) + self.P)

        return K.concatenate([sin_trans, original], -1)



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, tkr, value_miss
# Encoders Models
from .Seq2Seq.encoder_mdl import encoder_lstm
from .Seq2Seq.encoder_mdl import encoder_lstm2
from .Seq2Seq.encoder_mdl import encoder_batch2_lstm2_drop2
from .Seq2Seq.encoder_mdl import encoder_bilstm_state
from .Seq2Seq.encoder_mdl import encoder_bilstm2
from .Seq2Seq.encoder_mdl import encoder_conv1d
# Decoders Models
from .Seq2Seq.decoder_mdl import decoder_lstm
from .Seq2Seq.decoder_mdl import decoder_lstm2
from .Seq2Seq.decoder_mdl import decoder_lstm2_bach_drop
from .Seq2Seq.decoder_mdl import decoder_bilstm
from .Seq2Seq.decoder_mdl import decoder_bilstm2
from .Seq2Seq.decoder_mdl import decoder_bilstm_time_dist


class seq2seq_lstm(tkm.Model):
    def __init__(self, n_future=None, n_out_ft=None, config=None):
        super(seq2seq_lstm, self).__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.encoder = encoder_lstm(n_future)
        self.decoder = decoder_lstm(n_out_ft)

    def call(self, inputs):
        seq2seq = self.mask(inputs)
        seq2seq = self.encoder(seq2seq)
        seq2seq = self.decoder(seq2seq)
        return seq2seq


class seq2seq_lstm2(tkm.Model):
    def __init__(self, n_future=None, n_out_ft=None, config=None):
        super(seq2seq_lstm2, self).__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.encoder = encoder_lstm2(n_future)
        self.decoder = decoder_lstm2(n_out_ft)

    def call(self, inputs):
        seq2seq = self.mask(inputs)
        seq2seq = self.encoder(seq2seq)
        seq2seq = self.decoder(seq2seq)
        return seq2seq


class seq2seq_lstm_batch_drop(tkm.Model):

    def __init__(self, n_future=None, n_out_ft=None, config=None):
        super(seq2seq_lstm_batch_drop, self).__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.encoder = encoder_batch2_lstm2_drop2(n_future)
        self.decoder = decoder_lstm2_bach_drop(n_out_ft)

    def call(self, inputs):
        seq2seq = self.mask(inputs)
        seq2seq = self.encoder(seq2seq)
        seq2seq = self.decoder(seq2seq)
        return seq2seq


class seq2seq_bilstm(tkm.Model):
    def __init__(self, n_future=None, n_out_ft=None, config=None):
        super(seq2seq_bilstm, self).__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.encoder = encoder_bilstm_state(n_future)
        self.decoder = decoder_bilstm(n_out_ft)

    def call(self, inputs):
        seq2seq = self.mask(inputs)
        seq2seq = self.encoder(seq2seq)
        seq2seq = self.decoder(seq2seq)
        return seq2seq


class seq2seq_bilstm2(tkm.Model):
    def __init__(self, n_future=None, n_out_ft=None, config=None):
        super(seq2seq_bilstm2, self).__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.encoder = encoder_bilstm2(n_future)
        self.decoder = decoder_bilstm2(n_out_ft)

    def call(self, inputs):
        seq2seq = self.mask(inputs)
        seq2seq = self.encoder(seq2seq)
        seq2seq = self.decoder(seq2seq)
        return seq2seq


class seq2seq_conv1d_bilstm(tkm.Model):
    def __init__(self, n_future=None, n_out_ft=None, config=None):
        super(seq2seq_conv1d_bilstm, self).__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.encoder = encoder_conv1d(n_future)
        self.decoder = decoder_bilstm_time_dist(n_out_ft)

    def call(self, inputs):
        seq2seq = self.mask(inputs)
        seq2seq = self.encoder(seq2seq)
        seq2seq = self.decoder(seq2seq)
        return seq2seq


'''
class seq2seq_multihead_conv1d_bilstm(tkm.Model):
    def __init__(self, n_future=None, n_out_ft=None, n_inp_ft=None, config=None):
        super(seq2seq_multihead_conv1d_bilstm, self).__init__()
        self.encoder = _encoder_multihead_conv1d_lstm(n_future, n_inp_ft)
        self.decoder = _decoder_bilstm_time_dist(n_out_ft)

    def call(self, inputs):
        model = self.encoder(inputs)
        model = self.decoder(model)
        return model
'''


def seq2seq_multihead_conv1d_bilstm(n_past, n_inp_ft, n_future, n_out_ft, units=128, dropout=0.2):
    input_layer = tkl.Input(shape=(n_past, n_inp_ft))
    head_list = []
    for i in range(0, n_inp_ft):
        mask = tkl.Masking(mask_value=value_miss)(input_layer)
        conv_layer_head = tkl.Conv1D(
            filters=64, kernel_size=3, activation='relu', padding='causal', dilation_rate=2
        )(mask)
        conv_layer_head = tkl.Conv1D(filters=64, kernel_size=3, activation='relu')(conv_layer_head)
        conv_layer_flatten = tkl.Flatten()(conv_layer_head)
        head_list.append(conv_layer_flatten)
    concat_cnn = tkl.Concatenate(axis=1)(head_list)
    reshape = tkl.Reshape((head_list[0].shape[1], n_inp_ft))(concat_cnn)
    lstm = tkl.LSTM(units, activation='relu')(reshape)
    repeat = tkl.RepeatVector(n_future)(lstm)
    lstm_2 = tkl.LSTM(units, activation='relu', return_sequences=True)(repeat)
    dropout = tkl.Dropout(dropout)(lstm_2)
    dense = tkl.Dense(n_out_ft, activation='linear')(dropout)
    model = tkm.Model(inputs=input_layer, outputs=dense)
    return model


def seq2seq_bilstm_with_attention(n_past, n_inp_ft, n_future, n_out_ft, units=128, dropout=0.2):
    enc_inputs = tkl.Input(shape=(n_past, n_inp_ft))
    mask = tkl.Masking(mask_value=value_miss)(enc_inputs)
    # ENCODER
    enc_stack_h, enc_last_h_fw, enc_last_c_fw, enc_last_h_bw, enc_last_c_bw = tkl.Bidirectional(
        tkl.LSTM(
            units, dropout=dropout, return_state=True, return_sequences=True
        )
    )(mask)
    # encoder_stack_h.shape=(None, 450, 256)
    encoder_last_h = tkl.concatenate([enc_last_h_fw, enc_last_h_bw])
    encoder_last_c = tkl.concatenate([enc_last_c_fw, enc_last_c_bw])
    # encoder_last_h.shape=(None, 256) ---> last hidden_state
    # encoder_last_c.shape(None, 256) ---> last cell state
    # DECODER
    dec_inputs = tkl.RepeatVector(n_future)(encoder_last_h)
    dec_stack_h = tkl.Bidirectional(
        tkl.LSTM(
            units, dropout=dropout, return_sequences=True
        )
    )(dec_inputs, initial_state=[enc_last_h_fw, enc_last_c_fw, enc_last_h_bw, enc_last_c_bw])
    # ATTENTION LAYER
    attention = tkl.dot([dec_stack_h, enc_stack_h], axes=[2, 2])
    attention = tkl.Activation('softmax')(attention)
    context = tkl.dot([attention, enc_stack_h], axes=[2, 1])
    dec_combined_context = tkl.concatenate([context, dec_stack_h])
    out = tkl.TimeDistributed(tkl.Dense(n_out_ft))(dec_combined_context)
    model = tkm.Model(inputs=enc_inputs, outputs=out, name='model')
    return model


def seq2seq_lstm_with_loung_attention(n_past, n_inp_ft, n_future, n_out_ft, units=128, dropout=0.2):
    input_train = tkl.Input(shape=(n_past, n_inp_ft))
    mask = tkl.Masking(mask_value=value_miss)(input_train)
    output_train = tkl.Input(shape=(n_future, n_out_ft))

    encoder_stack_h, encoder_last_h, encoder_last_c = tkl.LSTM(
        units,
        activation='elu',
        dropout=dropout,
        recurrent_dropout=0.2,
        return_state=True,
        return_sequences=True,
        kernel_regularizer=tkr.l2(0.01),
        recurrent_regularizer=tkr.l2(0.01),
        bias_regularizer=tkr.l2(0.01)
    )(mask)
    encoder_last_h = tkl.BatchNormalization(momentum=0.6)(encoder_last_h)
    encoder_last_c = tkl.BatchNormalization(momentum=0.6)(encoder_last_c)
    # Repeat Vector
    decoder_input = tkl.RepeatVector(output_train.shape[1])(encoder_last_h)
    decoder_stack_h = tkl.LSTM(
        units,
        activation='elu',
        dropout=dropout,
        recurrent_dropout=0.2,
        return_state=False,
        return_sequences=True,
        kernel_regularizer=tkr.l2(0.01),
        recurrent_regularizer=tkr.l2(0.01),
        bias_regularizer=tkr.l2(0.01)
    )(decoder_input, initial_state=[encoder_last_h, encoder_last_c])
    attention = tkl.dot([decoder_stack_h, encoder_stack_h], axes=[2, 2])
    attention = tkl.Activation('softmax')(attention)
    context = tkl.dot([attention, encoder_stack_h], axes=[2, 1])
    context = tkl.BatchNormalization(momentum=0.6)(context)
    decoder_combined_context = tkl.concatenate([context, decoder_stack_h])
    out = tkl.TimeDistributed(tkl.Dense(output_train.shape[2]))(decoder_combined_context)
    built_model = tkm.Model(inputs=input_train, outputs=out, name='Seq2Seq_LSTM_with_Attention')
    return built_model



# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd

value_miss = -1
seed_lab = 44


class init_class:

    def __init__(self, dataframe=None):
        self.dataframe = dataframe


class init_models:

    def __init__(
            self,
            df_miss: pd.DataFrame = None,
            df_true: pd.DataFrame = None
    ):
        self.df_miss = df_miss
        self.df_true = df_true



# -*- coding: utf-8 -*-
# @Time    : 01/08/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

class base_arch:

    def __init__(
            self,
            name_mdl,
            type_mdl,
            n_past,
            n_future,
            n_inp_ft,
            n_out_ft,
            config_arch
    ):
        self.name_mdl = name_mdl
        self.type_mdl = type_mdl
        self.n_past = n_past
        self.n_future = n_future
        self.n_inp_ft = n_inp_ft
        self.n_out_ft = n_out_ft
        self.config_arch = config_arch



from tcn import TCN
from openpy_fxts.architectures.layers.layers_class import dense_multi_out, time2vec_birnn_layer, multi_time2vec_birnn, type_rnn
from openpy_fxts.base_tf import tkl, tkm


class tcn_bilstm(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(tcn_bilstm, self).__init__()
        if config is None:
            config = {
                'dense': {
                    'activation': 'linear'
                }
            }
        self.tcn = TCN(
            nb_filters=32,
            kernel_size=16,
            nb_stacks=4,
            dilations=(2, 4, 8, 16, 32, 64),
            padding='causal',
            use_skip_connections=True,
            dropout_rate=0.0,
            activation='tanh',
            kernel_initializer='glorot_uniform',
            use_batch_norm=False,
            use_layer_norm=False,
            use_weight_norm=False,
            return_sequences=True
        )
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=n_out_ft,
                return_sequences=False,
                activation='tanh'
            )
        )
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.tcn(inputs)
        x = self.bilstm(x)
        x = self.dense_out(x)
        return x


class time2vec_bilstm(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(time2vec_bilstm, self).__init__()
        if config is None:
            config = _get_config_mlp('time2vec_bilstm')
        self.hidden_layers = multi_time2vec_birnn(
            mdl=type_rnn(config['rnn_layers']['type']),
            output_dims=config['time2vec']['units'],
            units=config['rnn_layers']['units'],
            rnn_acts=config['rnn_layers']['activations'],
            return_seq=config['rnn_layers']['sequences'],
            is_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
        )
        self.hidden_time2vec_birnn = time2vec_birnn_layer(
            mdl=type_rnn(config['rnn_layers']['type']),
            output_dim=config['time2vec']['units'][-1],
            unit=n_out_ft,
            rnn_act='tanh',
            rnn_seq=False,
            is_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
        )
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.hidden_layers(inputs)
        x = self.hidden_time2vec_birnn(x)
        x = self.dense_out(x)
        return x